from mapproxy.grid import _create_tile_list, GridError


class MetaGrid(object):
    """
    This class contains methods to calculate bbox, etc. of metatiles.

    :param grid: the grid to use for the metatiles
    :param meta_size: the number of tiles a metatile consist
    :type meta_size: ``(x_size, y_size)``
    :param meta_buffer: the buffer size in pixel that is added to each metatile.
        the number is added to all four borders.
        this buffer may improve the handling of lables overlapping (meta)tile borders.
    :type meta_buffer: pixel
    """

    def __init__(self, grid, meta_size, meta_buffer=0):
        self.grid = grid
        self.meta_size = meta_size or 0
        self.meta_buffer = meta_buffer

    def _meta_bbox(self, tile_coord=None, tiles=None, limit_to_bbox=True):
        """
        Returns the bbox of the metatile that contains `tile_coord`.

        :type tile_coord: ``(x, y, z)``

        >>> from mapproxy.grid.tile_grid import TileGrid
        >>> mgrid = MetaGrid(grid=TileGrid(), meta_size=(2, 2))
        >>> [round(x, 2) for x in mgrid._meta_bbox((0, 0, 2))[0]]
        [-20037508.34, -20037508.34, 0.0, 0.0]
        >>> mgrid = MetaGrid(grid=TileGrid(), meta_size=(2, 2))
        >>> [round(x, 2) for x in mgrid._meta_bbox((0, 0, 0))[0]]
        [-20037508.34, -20037508.34, 20037508.34, 20037508.34]
        """
        if tiles:
            assert tile_coord is None
            level = tiles[0][2]
            bbox = self.grid._tiles_bbox(tiles)
        else:
            level = tile_coord[2]
            bbox = self.unbuffered_meta_bbox(tile_coord)
        return self._buffered_bbox(bbox, level, limit_to_bbox)

    def unbuffered_meta_bbox(self, tile_coord):
        x, y, z = tile_coord

        meta_size = self._meta_size(z)

        return self.grid._tiles_bbox([(tile_coord),
                                      (x+meta_size[0]-1, y+meta_size[1]-1, z)])

    def _buffered_bbox(self, bbox, level, limit_to_grid_bbox=True):
        minx, miny, maxx, maxy = bbox

        buffers = (0, 0, 0, 0)
        if self.meta_buffer > 0:
            res = self.grid.resolution(level)
            minx -= self.meta_buffer * res
            miny -= self.meta_buffer * res
            maxx += self.meta_buffer * res
            maxy += self.meta_buffer * res
            buffers = [self.meta_buffer, self.meta_buffer, self.meta_buffer, self.meta_buffer]

            if limit_to_grid_bbox:
                if self.grid.bbox[0] > minx:
                    delta = self.grid.bbox[0] - minx
                    buffers[0] = buffers[0] - int(round(delta / res, 5))
                    minx = self.grid.bbox[0]
                if self.grid.bbox[1] > miny:
                    delta = self.grid.bbox[1] - miny
                    buffers[1] = buffers[1] - int(round(delta / res, 5))
                    miny = self.grid.bbox[1]
                if self.grid.bbox[2] < maxx:
                    delta = maxx - self.grid.bbox[2]
                    buffers[2] = buffers[2] - int(round(delta / res, 5))
                    maxx = self.grid.bbox[2]
                if self.grid.bbox[3] < maxy:
                    delta = maxy - self.grid.bbox[3]
                    buffers[3] = buffers[3] - int(round(delta / res, 5))
                    maxy = self.grid.bbox[3]
        return (minx, miny, maxx, maxy), tuple(buffers)

    def meta_tile(self, tile_coord):
        """
        Returns the meta tile for `tile_coord`.
        """
        tile_coord = self.main_tile(tile_coord)
        level = tile_coord[2]
        bbox, buffers = self._meta_bbox(tile_coord)
        grid_size = self._meta_size(level)
        size = self._size_from_buffered_bbox(bbox, level)

        tile_patterns = self._tiles_pattern(tile=tile_coord, grid_size=grid_size, buffers=buffers)

        return MetaTile(bbox=bbox, size=size, tile_patterns=tile_patterns,
                        grid_size=grid_size
                        )

    def minimal_meta_tile(self, tiles):
        """
        Returns a MetaTile that contains all `tiles` plus ``meta_buffer``,
        but nothing more.
        """

        tiles, grid_size, bounds = self._full_tile_list(tiles)
        tiles = list(tiles)
        bbox, buffers = self._meta_bbox(tiles=bounds)

        level = tiles[0][2]
        size = self._size_from_buffered_bbox(bbox, level)

        tile_pattern = self._tiles_pattern(tiles=tiles, grid_size=grid_size, buffers=buffers)

        return MetaTile(
            bbox=bbox,
            size=size,
            tile_patterns=tile_pattern,
            grid_size=grid_size,
        )

    def _size_from_buffered_bbox(self, bbox, level):
        # meta_size * tile_size + 2*buffer does not work,
        # since the buffer can get truncated at the grid border
        res = self.grid.resolution(level)
        width = int(round((bbox[2] - bbox[0]) / res))
        height = int(round((bbox[3] - bbox[1]) / res))
        return width, height

    def _full_tile_list(self, tiles):
        """
        Return a complete list of all tiles that a minimal meta tile with `tiles` contains.

        >>> from mapproxy.grid.tile_grid import TileGrid
        >>> mgrid = MetaGrid(grid=TileGrid(), meta_size=(2, 2))
        >>> mgrid._full_tile_list([(0, 0, 2), (1, 1, 2)])
        ([(0, 1, 2), (1, 1, 2), (0, 0, 2), (1, 0, 2)], (2, 2), ((0, 0, 2), (1, 1, 2)))
        """
        tile = tiles.pop()
        z = tile[2]
        minx = maxx = tile[0]
        miny = maxy = tile[1]

        for tile in tiles:
            x, y = tile[:2]
            minx = min(minx, x)
            maxx = max(maxx, x)
            miny = min(miny, y)
            maxy = max(maxy, y)

        grid_size = 1+maxx-minx, 1+maxy-miny

        if self.grid.flipped_y_axis:
            ys = range(miny, maxy+1)
        else:
            ys = range(maxy, miny-1, -1)
        xs = range(minx, maxx+1)

        bounds = (minx, miny, z), (maxx, maxy, z)

        return list(_create_tile_list(xs, ys, z, (maxx+1, maxy+1))), grid_size, bounds

    def main_tile(self, tile_coord):
        x, y, z = tile_coord

        meta_size = self._meta_size(z)

        x0 = x//meta_size[0] * meta_size[0]
        y0 = y//meta_size[1] * meta_size[1]

        return x0, y0, z

    def tile_list(self, main_tile):
        tile_grid = self._meta_size(main_tile[2])
        return self._meta_tile_list(main_tile, tile_grid)

    def _meta_tile_list(self, main_tile, tile_grid):
        """
        >>> from mapproxy.grid.tile_grid import TileGrid
        >>> mgrid = MetaGrid(grid=TileGrid(), meta_size=(2, 2))
        >>> mgrid._meta_tile_list((0, 1, 3), (2, 2))
        [(0, 1, 3), (1, 1, 3), (0, 0, 3), (1, 0, 3)]
        """
        minx, miny, z = self.main_tile(main_tile)
        maxx = minx + tile_grid[0] - 1
        maxy = miny + tile_grid[1] - 1
        if self.grid.flipped_y_axis:
            ys = range(miny, maxy+1)
        else:
            ys = range(maxy, miny-1, -1)
        xs = range(minx, maxx+1)

        return list(_create_tile_list(xs, ys, z, self.grid.grid_sizes[z]))

    def _tiles_pattern(self, grid_size, buffers, tile=None, tiles=None):
        """
        Returns the tile pattern for the given list of tiles.
        The result contains for each tile the ``tile_coord`` and the upper-left
        pixel coordinate of the tile in the meta tile image.

        >>> from mapproxy.grid.tile_grid import TileGrid
        >>> mgrid = MetaGrid(grid=TileGrid(), meta_size=(2, 2))
        >>> tiles = list(mgrid._tiles_pattern(tiles=[(0, 1, 2), (1, 1, 2)],
        ...                                   grid_size=(2, 1),
        ...                                   buffers=(0, 0, 10, 10)))
        >>> tiles[0], tiles[-1]
        (((0, 1, 2), (0, 10)), ((1, 1, 2), (256, 10)))

        >>> tiles = list(mgrid._tiles_pattern(tile=(1, 1, 2),
        ...                                   grid_size=(2, 2),
        ...                                   buffers=(10, 20, 30, 40)))
        >>> tiles[0], tiles[-1]
        (((0, 1, 2), (10, 40)), ((1, 0, 2), (266, 296)))

        """
        if tile:
            tiles = self._meta_tile_list(tile, grid_size)

        for i in range(grid_size[1]):
            for j in range(grid_size[0]):
                yield tiles[j+i*grid_size[0]], (
                    j*self.grid.tile_size[0] + buffers[0],
                    i*self.grid.tile_size[1] + buffers[3])

    def _meta_size(self, level):
        grid_size = self.grid.grid_sizes[level]
        return min(self.meta_size[0], grid_size[0]), min(self.meta_size[1], grid_size[1])

    def get_affected_level_tiles(self, bbox, level):
        """
        Get a list with all affected tiles for a `bbox` in the given `level`.

        :returns: the bbox, the size and a list with tile coordinates, sorted row-wise
        :rtype: ``bbox, (xs, yz), [(x, y, z), ...]``

        >>> from mapproxy.grid.tile_grid import TileGrid
        >>> grid = MetaGrid(TileGrid(), (2, 2))
        >>> bbox = (-20037508.34, -20037508.34, 20037508.34, 20037508.34)
        >>> grid.get_affected_level_tiles(bbox, 0)
        ... #doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        ((-20037508.342789244, -20037508.342789244,\
          20037508.342789244, 20037508.342789244), (1, 1),\
          <generator object ...>)
        """

        # remove 1/10 of a pixel so we don't get a tiles we only touch
        delta = self.grid.resolutions[level] / 10.0
        x0, y0, _ = self.grid.tile(bbox[0]+delta, bbox[1]+delta, level)
        x1, y1, _ = self.grid.tile(bbox[2]-delta, bbox[3]-delta, level)

        meta_size = self._meta_size(level)

        x0 = x0//meta_size[0] * meta_size[0]
        x1 = x1//meta_size[0] * meta_size[0]
        y0 = y0//meta_size[1] * meta_size[1]
        y1 = y1//meta_size[1] * meta_size[1]

        try:
            return self._tile_iter(x0, y0, x1, y1, level)
        except IndexError:
            raise GridError('Invalid BBOX')

    def _tile_iter(self, x0, y0, x1, y1, level):
        meta_size = self._meta_size(level)

        xs = list(range(x0, x1+1, meta_size[0]))
        if self.grid.flipped_y_axis:
            y0, y1 = y1, y0
            ys = list(range(y0, y1+1, meta_size[1]))
        else:
            ys = list(range(y1, y0-1, -meta_size[1]))

        ll = (xs[0], ys[-1], level)
        ur = (xs[-1], ys[0], level)
        # add meta_size to get full affected bbox
        ur = ur[0]+meta_size[0]-1, ur[1]+meta_size[1]-1, ur[2]
        abbox = self.grid._tiles_bbox([ll, ur])
        return (abbox, (len(xs), len(ys)),
                _create_tile_list(xs, ys, level, self.grid.grid_sizes[level]))


class MetaTile(object):
    def __init__(self, bbox, size, tile_patterns, grid_size):
        self.bbox = bbox
        self.size = size
        self.tile_patterns = list(tile_patterns)
        self.grid_size = grid_size

    @property
    def tiles(self):
        return [t[0] for t in self.tile_patterns]

    @property
    def main_tile_coord(self):
        """
        Returns the "main" tile of the meta tile. This tile(coord) can be used
        for locking.

        >>> t = MetaTile(None, None, [((0, 0, 0), (0, 0)), ((1, 0, 0), (100, 0))], (2, 1))
        >>> t.main_tile_coord
        (0, 0, 0)
        >>> t = MetaTile(None, None, [(None, None), ((1, 0, 0), (100, 0))], (2, 1))
        >>> t.main_tile_coord
        (1, 0, 0)
        """
        for t in self.tiles:
            if t is not None:
                return t

    def __repr__(self):
        return "MetaTile(%r, %r, %r, %r)" % (self.bbox, self.size, self.grid_size,
                                             self.tile_patterns)



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from mapproxy.image.opts import (
    ImageOptions,
    create_image,
    compatible_image_options,
)


class TestCreateImage(object):

    def test_default(self):
        img = create_image((100, 100))
        assert img.size == (100, 100)
        assert img.mode == "RGB"
        assert img.getcolors() == [(100 * 100, (255, 255, 255))]

    def test_transparent(self):
        img = create_image((100, 100), ImageOptions(transparent=True))
        assert img.size == (100, 100)
        assert img.mode == "RGBA"
        assert img.getcolors() == [(100 * 100, (255, 255, 255, 0))]

    def test_transparent_rgb(self):
        img = create_image(
            (100, 100), ImageOptions(mode="RGB", transparent=True)
        )
        assert img.size == (100, 100)
        assert img.mode == "RGB"
        assert img.getcolors() == [(100 * 100, (255, 255, 255))]

    def test_bgcolor(self):
        img = create_image((100, 100), ImageOptions(bgcolor=(200, 100, 0)))
        assert img.size == (100, 100)
        assert img.mode == "RGB"
        assert img.getcolors() == [(100 * 100, (200, 100, 0))]

    def test_rgba_bgcolor(self):
        img = create_image((100, 100), ImageOptions(bgcolor=(200, 100, 0, 30)))
        assert img.size == (100, 100)
        assert img.mode == "RGB"
        assert img.getcolors() == [(100 * 100, (200, 100, 0))]

    def test_rgba_bgcolor_transparent(self):
        img = create_image(
            (100, 100),
            ImageOptions(bgcolor=(200, 100, 0, 30), transparent=True),
        )
        assert img.size == (100, 100)
        assert img.mode == "RGBA"
        assert img.getcolors() == [(100 * 100, (200, 100, 0, 30))]

    def test_rgba_bgcolor_rgba_mode(self):
        img = create_image(
            (100, 100), ImageOptions(bgcolor=(200, 100, 0, 30), mode="RGBA")
        )
        assert img.size == (100, 100)
        assert img.mode == "RGBA"
        assert img.getcolors() == [(100 * 100, (200, 100, 0, 30))]


class TestCompatibleImageOptions(object):

    def test_formats(self):
        img_opts = compatible_image_options(
            [
                ImageOptions(format="image/png"),
                ImageOptions(format="image/jpeg"),
            ]
        )
        assert img_opts.format == "image/png"

        img_opts = compatible_image_options(
            [
                ImageOptions(format="image/png"),
                ImageOptions(format="image/jpeg"),
            ],
            ImageOptions(format="image/tiff"),
        )
        assert img_opts.format == "image/tiff"

    def test_colors(self):
        img_opts = compatible_image_options(
            [ImageOptions(colors=None), ImageOptions(colors=16)]
        )
        assert img_opts.colors == 16

        img_opts = compatible_image_options(
            [ImageOptions(colors=256), ImageOptions(colors=16)]
        )
        assert img_opts.colors == 256

        img_opts = compatible_image_options(
            [ImageOptions(colors=256), ImageOptions(colors=16)],
            ImageOptions(colors=4),
        )
        assert img_opts.colors == 4

        img_opts = compatible_image_options(
            [ImageOptions(colors=256), ImageOptions(colors=0)]
        )
        assert img_opts.colors == 0

    def test_transparent(self):
        img_opts = compatible_image_options(
            [ImageOptions(transparent=False), ImageOptions(transparent=True)]
        )
        assert not img_opts.transparent

        img_opts = compatible_image_options(
            [ImageOptions(transparent=None), ImageOptions(transparent=True)]
        )
        assert img_opts.transparent

        img_opts = compatible_image_options(
            [ImageOptions(transparent=None), ImageOptions(transparent=True)],
            ImageOptions(transparent=None),
        )
        assert img_opts.transparent

        img_opts = compatible_image_options(
            [ImageOptions(transparent=True), ImageOptions(transparent=True)]
        )
        assert img_opts.transparent

    def test_mode(self):
        img_opts = compatible_image_options(
            [ImageOptions(mode="RGB"), ImageOptions(mode="P")]
        )
        assert img_opts.mode == "RGB"

        img_opts = compatible_image_options(
            [ImageOptions(mode="RGBA"), ImageOptions(mode="P")]
        )
        assert img_opts.mode == "RGBA"

        img_opts = compatible_image_options(
            [ImageOptions(mode="RGB"), ImageOptions(mode="P")]
        )
        assert img_opts.mode == "RGB"

        img_opts = compatible_image_options(
            [ImageOptions(mode="RGB"), ImageOptions(mode="P")],
            ImageOptions(mode="P"),
        )
        assert img_opts.mode == "P"



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile

from mapproxy.util.yaml import load_yaml, load_yaml_file, YAMLError


class TestLoadYAMLFile(object):

    def setup_method(self):
        self.tmp_files = []

    def teardown_method(self):
        for f in self.tmp_files:
            os.unlink(f)

    def yaml_file(self, content):
        fd, fname = tempfile.mkstemp()
        f = os.fdopen(fd, "w")
        f.write(content)
        self.tmp_files.append(fname)
        return fname

    def test_load_yaml_file(self):
        f = self.yaml_file("hello:\n - 1\n - 2")
        doc = load_yaml_file(open(f))
        assert doc == {"hello": [1, 2]}

    def test_load_yaml_file_filename(self):
        f = self.yaml_file("hello:\n - 1\n - 2")
        assert isinstance(f, str)
        doc = load_yaml_file(f)
        assert doc == {"hello": [1, 2]}

    def test_load_yaml(self):
        doc = load_yaml("hello:\n - 1\n - 2")
        assert doc == {"hello": [1, 2]}

    def test_load_yaml_with_tabs(self):
        try:
            f = self.yaml_file("hello:\n\t- world")
            load_yaml_file(f)
        except YAMLError as ex:
            assert "line 2" in str(ex)
        else:
            assert False, "expected YAMLError"

    def test_load_yaml_string_error(self):
        try:
            load_yaml('only a string')
        except YAMLError as ex:
            assert "not a YAML dict" in str(ex)
        else:
            assert False, "expected YAMLError"



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import time

from contextlib import contextmanager

from mapproxy.util.lock import FileLock, cleanup_lockdir, DummyLock


class CacheBackendError(Exception):
    pass


@contextmanager
def tile_buffer(tile):
    data = tile.source.as_buffer(seekable=True)
    data.seek(0)
    yield data
    tile.size = data.tell()
    if not tile.timestamp:
        tile.timestamp = time.time()
    data.seek(0)
    tile.stored = True


class TileCacheBase(object):
    """
    Base implementation of a tile cache.
    """

    supports_timestamp = True
    supports_dimensions = False

    def __init__(self, coverage=None) -> None:
        self.coverage = coverage

    def load_tile(self, tile, with_metadata=False, dimensions=None):
        raise NotImplementedError()

    def load_tiles(self, tiles, with_metadata=False, dimensions=None):
        all_succeed = True
        for tile in tiles:
            if not self.load_tile(tile, with_metadata=with_metadata, dimensions=dimensions):
                all_succeed = False
        return all_succeed

    def store_tile(self, tile, dimensions=None):
        raise NotImplementedError()

    def store_tiles(self, tiles, dimensions=None):
        all_succeed = True
        for tile in tiles:
            if not self.store_tile(tile, dimensions=dimensions):
                all_succeed = False
        return all_succeed

    def remove_tile(self, tile, dimensions=None):
        raise NotImplementedError()

    def remove_tiles(self, tiles, dimensions=None):
        for tile in tiles:
            self.remove_tile(tile, dimensions=dimensions)

    def is_cached(self, tile, dimensions=None):
        """
        Return ``True`` if the tile is cached.
        """
        raise NotImplementedError()

    def load_tile_metadata(self, tile, dimensions=None):
        """
        Fill the metadata attributes of `tile`.
        Sets ``.timestamp`` and ``.size``.
        """
        raise NotImplementedError()


# whether we immediately remove lock files or not
REMOVE_ON_UNLOCK = True
if sys.platform == 'win32':
    # windows does not handle this well
    REMOVE_ON_UNLOCK = False


class TileLocker(object):
    def __init__(self, lock_dir, lock_timeout, lock_cache_id, directory_permissions=None, file_permissions=None):
        self.lock_dir = lock_dir
        self.lock_timeout = lock_timeout
        self.lock_cache_id = lock_cache_id
        self.directory_permissions = directory_permissions
        self.file_permissions = file_permissions

    def lock_filename(self, tile):
        return os.path.join(self.lock_dir, self.lock_cache_id + '-' +
                            '-'.join(map(str, tile.coord)) + '.lck')

    def lock(self, tile):
        """
        Returns a lock object for this tile.
        """
        if getattr(self, 'locking_disabled', False):
            return DummyLock()
        lock_filename = self.lock_filename(tile)
        cleanup_lockdir(self.lock_dir, max_lock_time=self.lock_timeout + 10,
                        force=False)
        return FileLock(lock_filename, timeout=self.lock_timeout,
                        remove_on_unlock=REMOVE_ON_UNLOCK, directory_permissions=self.directory_permissions,
                        file_permissions=self.file_permissions)



# This file is part of the MapProxy project.
# Copyright (C) 2010-2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import hashlib

from io import BytesIO

import pytest

from mapproxy.util.bbox import bbox_equals
from mapproxy.util.times import format_httpdate
from mapproxy.test.image import is_jpeg, tmp_image
from mapproxy.test.http import mock_httpd
from mapproxy.test.helper import validate_with_xsd
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "kml_layer.yaml"


ns = {"kml": "http://www.opengis.net/kml/2.2"}


class TestKML(SysTest):

    def test_get_out_of_bounds_tile(self, app):
        for coord in [(0, 0, -1), (-1, 0, 0), (0, -1, 0), (4, 2, 1), (1, 3, 0)]:
            x, y, z = coord
            url = "/kml/wms_cache/%d/%d/%d.kml" % (z, x, y)
            resp = app.get(url, status=404)
            assert "outside the bounding box" in resp

    def test_invalid_layer(self, app):
        resp = app.get("/kml/inVAlid/0/0/0.png", status=404)
        assert resp.content_type == "text/plain"
        assert "unknown layer: inVAlid" in resp

    def test_invalid_format(self, app):
        resp = app.get("/kml/wms_cache/0/0/1.png", status=404)
        assert resp.content_type == "text/plain"
        assert "invalid format" in resp

    def test_get_tile_tile_source_error(self, app):
        resp = app.get("/kml/wms_cache/0/0/0.jpeg", status=500)
        assert resp.content_type == "text/plain"
        assert "No response from URL" in resp

    def _check_tile_resp(self, resp):
        assert resp.content_type == "image/jpeg"
        assert resp.content_length == len(resp.body)
        data = BytesIO(resp.body)
        assert is_jpeg(data)

    def _update_timestamp(self, base_config, cache_dir):
        timestamp = 1234567890.0
        size = 10214
        os.utime(
            cache_dir.join(
                "wms_cache_EPSG900913/01/000/000/000/000/000/001.jpeg"
            ).strpath,
            (timestamp, timestamp),
        )
        max_age = base_config.tiles.expires_hours * 60 * 60
        md5 = hashlib.new('md5', (str(timestamp) + str(size)).encode("ascii"), usedforsecurity=False)
        etag = md5.hexdigest()
        return etag, max_age

    def _check_cache_control_headers(self, resp, etag, max_age, timestamp=1234567890.0):
        assert resp.headers["ETag"] == etag
        if timestamp is None:
            assert "Last-modified" not in resp.headers
        else:
            assert resp.headers["Last-modified"] == format_httpdate(timestamp)
        assert resp.headers["Cache-control"] == "public, max-age=%d, s-maxage=%d" % (
            max_age,
            max_age,
        )

    def test_get_cached_tile(self, app, base_config, cache_dir, fixture_cache_data):
        etag, max_age = self._update_timestamp(base_config, cache_dir)
        resp = app.get("/kml/wms_cache/1/0/1.jpeg")
        self._check_cache_control_headers(resp, etag, max_age)
        self._check_tile_resp(resp)

    def test_if_none_match(self, app, base_config, cache_dir, fixture_cache_data):
        etag, max_age = self._update_timestamp(base_config, cache_dir)
        resp = app.get("/kml/wms_cache/1/0/1.jpeg", headers={"If-None-Match": etag})
        assert resp.status == "304 Not Modified"
        self._check_cache_control_headers(resp, etag, max_age)

        resp = app.get(
            "/kml/wms_cache/1/0/1.jpeg", headers={"If-None-Match": etag + "foo"}
        )
        self._check_cache_control_headers(resp, etag, max_age)
        assert resp.status == "200 OK"
        self._check_tile_resp(resp)

    def test_get_kml(self, app, base_config):
        resp = app.get("/kml/wms_cache/0/0/0.kml")
        xml = resp.lxml
        assert validate_with_xsd(xml, "kml/2.2.0/ogckml22.xsd")
        assert bbox_equals(
            self._bbox(xml.xpath("/kml:kml/kml:Document", namespaces=ns)[0]),
            (-180, -90, 180, 90),
        )
        assert bbox_equals(
            self._bbox(
                xml.xpath("/kml:kml/kml:Document/kml:GroundOverlay", namespaces=ns)[0]
            ),
            (-180, 0, 0, 90),
        )
        assert xml.xpath(
            "/kml:kml/kml:Document/kml:GroundOverlay/kml:Icon/kml:href/text()",
            namespaces=ns,
        ) == [
            "http://localhost/kml/wms_cache/EPSG900913/1/0/1.jpeg",
            "http://localhost/kml/wms_cache/EPSG900913/1/1/1.jpeg",
            "http://localhost/kml/wms_cache/EPSG900913/1/0/0.jpeg",
            "http://localhost/kml/wms_cache/EPSG900913/1/1/0.jpeg",
        ]
        assert xml.xpath(
            "/kml:kml/kml:Document/kml:NetworkLink/kml:Link/kml:href/text()",
            namespaces=ns,
        ) == [
            "http://localhost/kml/wms_cache/EPSG900913/1/0/1.kml",
            "http://localhost/kml/wms_cache/EPSG900913/1/1/1.kml",
            "http://localhost/kml/wms_cache/EPSG900913/1/0/0.kml",
            "http://localhost/kml/wms_cache/EPSG900913/1/1/0.kml",
        ]

        md5 = hashlib.new('md5', resp.body, usedforsecurity=False)
        etag = md5.hexdigest()
        max_age = base_config.tiles.expires_hours * 60 * 60
        self._check_cache_control_headers(resp, etag, max_age, None)

        resp = app.get("/kml/wms_cache/0/0/0.kml", headers={"If-None-Match": etag})
        assert resp.status == "304 Not Modified"

    def test_get_kml_init(self, app):
        resp = app.get("/kml/wms_cache")
        xml = resp.lxml
        assert validate_with_xsd(xml, "kml/2.2.0/ogckml22.xsd")
        assert xml.xpath(
            "/kml:kml/kml:Document/kml:GroundOverlay/kml:Icon/kml:href/text()",
            namespaces=ns,
        ) == [
            "http://localhost/kml/wms_cache/EPSG900913/1/0/1.jpeg",
            "http://localhost/kml/wms_cache/EPSG900913/1/1/1.jpeg",
            "http://localhost/kml/wms_cache/EPSG900913/1/0/0.jpeg",
            "http://localhost/kml/wms_cache/EPSG900913/1/1/0.jpeg",
        ]
        assert xml.xpath(
            "/kml:kml/kml:Document/kml:NetworkLink/kml:Link/kml:href/text()",
            namespaces=ns,
        ) == [
            "http://localhost/kml/wms_cache/EPSG900913/1/0/1.kml",
            "http://localhost/kml/wms_cache/EPSG900913/1/1/1.kml",
            "http://localhost/kml/wms_cache/EPSG900913/1/0/0.kml",
            "http://localhost/kml/wms_cache/EPSG900913/1/1/0.kml",
        ]

    def test_get_kml_nw(self, app):
        resp = app.get("/kml/wms_cache_nw/1/0/0.kml")
        xml = resp.lxml

        assert validate_with_xsd(xml, "kml/2.2.0/ogckml22.xsd")

        assert bbox_equals(
            self._bbox(xml.xpath("/kml:kml/kml:Document", namespaces=ns)[0]),
            (-180, -90, 0, 0),
        )
        assert bbox_equals(
            self._bbox(
                xml.xpath("/kml:kml/kml:Document/kml:GroundOverlay", namespaces=ns)[0]
            ),
            (-180, -66.51326, -90, 0),
        )

        assert xml.xpath(
            "/kml:kml/kml:Document/kml:GroundOverlay/kml:Icon/kml:href/text()",
            namespaces=ns,
        ) == [
            "http://localhost/kml/wms_cache_nw/EPSG900913/2/0/1.jpeg",
            "http://localhost/kml/wms_cache_nw/EPSG900913/2/1/1.jpeg",
            "http://localhost/kml/wms_cache_nw/EPSG900913/2/0/0.jpeg",
            "http://localhost/kml/wms_cache_nw/EPSG900913/2/1/0.jpeg",
        ]
        assert xml.xpath(
            "/kml:kml/kml:Document/kml:NetworkLink/kml:Link/kml:href/text()",
            namespaces=ns,
        ) == [
            "http://localhost/kml/wms_cache_nw/EPSG900913/2/0/1.kml",
            "http://localhost/kml/wms_cache_nw/EPSG900913/2/1/1.kml",
            "http://localhost/kml/wms_cache_nw/EPSG900913/2/0/0.kml",
            "http://localhost/kml/wms_cache_nw/EPSG900913/2/1/0.kml",
        ]

    def test_get_kml2(self, app):
        resp = app.get("/kml/wms_cache/1/0/1.kml")
        xml = resp.lxml
        assert validate_with_xsd(xml, "kml/2.2.0/ogckml22.xsd")

    def test_get_kml_multi_layer(self, app):
        resp = app.get("/kml/wms_cache_multi/1/0/0.kml")
        xml = resp.lxml
        assert validate_with_xsd(xml, "kml/2.2.0/ogckml22.xsd")
        assert xml.xpath(
            "/kml:kml/kml:Document/kml:GroundOverlay/kml:Icon/kml:href/text()",
            namespaces=ns,
        ) == [
            "http://localhost/kml/wms_cache_multi/EPSG4326/2/0/1.jpeg",
            "http://localhost/kml/wms_cache_multi/EPSG4326/2/1/1.jpeg",
            "http://localhost/kml/wms_cache_multi/EPSG4326/2/0/0.jpeg",
            "http://localhost/kml/wms_cache_multi/EPSG4326/2/1/0.jpeg",
        ]
        assert xml.xpath(
            "/kml:kml/kml:Document/kml:NetworkLink/kml:Link/kml:href/text()",
            namespaces=ns,
        ) == [
            "http://localhost/kml/wms_cache_multi/EPSG4326/2/0/1.kml",
            "http://localhost/kml/wms_cache_multi/EPSG4326/2/1/1.kml",
            "http://localhost/kml/wms_cache_multi/EPSG4326/2/0/0.kml",
            "http://localhost/kml/wms_cache_multi/EPSG4326/2/1/0.kml",
        ]

    def test_get_tile(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.1&BBOX=-20037508.3428,-20037508.3428,0.0,0.0"
                    "&WIDTH=256"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/kml/wms_cache/1/0/0.jpeg")
                assert resp.content_type == "image/jpeg"
        assert cache_dir.join(
            "wms_cache_EPSG900913/01/000/000/000/000/000/000.jpeg"
        ).check()

    def _bbox(self, elem):
        elems = elem.xpath("kml:Region/kml:LatLonAltBox", namespaces=ns)[0]
        n, s, e, w = [float(elem.text) for elem in elems.getchildren()]
        return w, s, e, n



# This file is part of the MapProxy project.
# Copyright (C) 2025 Spatialys
#
# Initial development funded by Centre National d'Etudes Spatiales (CNES): https://cnes.fr
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
import copy
import re
from typing import Optional

from mapproxy.image import (
    bbox_position_in_image,
    SubImageSource,
    BlankImageSource,
    GeoReference,
)
from mapproxy.image.merge import LayerMerger
from mapproxy.image.message import attribution_image
from mapproxy.image.opts import ImageOptions
from mapproxy.layer import MapQuery, MapExtent
from mapproxy.response import Response
from mapproxy.request.base import Request
from mapproxy.service.ogcapi.constants import FORMAT_TYPES, F_PNG, F_JPEG
from mapproxy.service.ogcapi.map_utils import (
    compute_width_height_bbox,
    get_aspect_ratio,
    get_width_height_from_default_size,
)
from mapproxy.service.ogcapi.server import OGCAPIServer, get_image_format
from mapproxy.service.wms import LayerRenderer, WMSLayer, WMSGroupLayer
from mapproxy.srs import SRS, ogc_crs_url_to_auth_code

# Cf https://www.w3.org/wiki/CSS/Properties/color/keywords
W3C_BGCOLORS = {
    "black": "#000000",
    "silver": "#C0C0C0",
    "gray": "#808080",
    "white": "#FFFFFF",
    "maroon": "#800000",
    "red": "#FF0000",
    "purple": "#800080",
    "fuchsia": "#FF00FF",
    "green": "#008000",
    "lime": "#00FF00",
    "olive": "#808000",
    "yellow": "#FFFF00",
    "navy": "#000080",
    "blue": "#0000FF",
    "teal": "#008080",
    "aqua": "#00FFFF",
}


def ogcapi_srs_to_srs(s: str):
    """Convert a OGC SRS, expressed either as a OGC CRS URL, "
    a safe CURIE ("[EPSG:4326]") or a unsafe CURIE ("EPSG:4326") to a
    MapProxy SRS object.
    """

    if s.startswith("http"):
        return SRS(ogc_crs_url_to_auth_code(s))
    elif s and s[0] == "[" and s[-1] == "]":
        # Safe CURIE
        return SRS(s[1:-1])
    else:
        # Unsafe CURIE
        return SRS(s)


def subset_to_bbox(subset):
    """Convert the "subset" parameter into a bounding box (list of 4 float)."""

    if not isinstance(subset, list):
        subset = [subset]

    if len(subset) == 1:
        subset = subset[0].split(",")

    if len(subset) != 2:
        raise OGCAPIServer.invalid_parameter("subset must include 2 axis")

    x_axis_set = set(["lon", "long", "longitude", "x", "e", "easting"])
    y_axis_set = set(["lat", "latitude", "y", "n", "northing"])
    bbox = [None, None, None, None]
    for subset_part in subset:
        subset_part = subset_part.lower()
        pattern = re.compile("^(\\w+)\\(([^)]+)\\)$")
        match = pattern.match(subset_part)
        if match is None:
            raise OGCAPIServer.invalid_parameter(f"Invalid subset part {subset_part}")
        axis_name, range_part = match.groups()
        range_part = range_part.split(":")
        if len(range_part) != 2:
            raise OGCAPIServer.invalid_parameter(
                f"Only intervals are supported in subset part {subset_part}"
            )

        if axis_name in x_axis_set:
            idx_min = 0
            idx_max = 2
        elif axis_name in y_axis_set:
            idx_min = 1
            idx_max = 3
        else:
            raise OGCAPIServer.invalid_parameter(
                f"Unsupported axis name {axis_name} in subset part {subset_part}"
            )

        if bbox[idx_min] is not None:
            raise OGCAPIServer.invalid_parameter(
                f"Axis name {axis_name} has been specified in multiple subset parts"
            )

        try:
            bbox[idx_min] = float(range_part[0])
            bbox[idx_max] = float(range_part[1])
        except ValueError:
            raise OGCAPIServer.invalid_parameter(
                f"Non numeric value found in range of subset part {subset_part}"
            )

    return bbox


def get_bbox_center_or_subset_crs(crs, layer, param_name):
    """Return the MapProxy SRS object corresponding to a
    "bbox-center"/"center-crs"/"subset-crs" parameter.
    """
    if crs:
        try:
            return ogcapi_srs_to_srs(crs)
        except ValueError:
            raise OGCAPIServer.invalid_parameter(f"{param_name} is not a valid CRS")
    else:
        if layer.extent.srs.srs_code.startswith("EPSG:"):
            # Default CRS for an earth projection
            return SRS("OGC:CRS84")
        else:
            return layer.extent.srs.get_geographic_srs()


def get_headers(server: OGCAPIServer, query: MapQuery):
    headers = copy.copy(server.response_headers)
    if query.srs.srs_code != "EPSG:4326":
        headers["Content-Crs"] = "<" + query.srs.to_ogc_url() + ">"
    if query.srs.is_axis_order_ne and query.srs.srs_code != "EPSG:4326":
        headers["Content-Bbox"] = "%.17g,%.17g,%.17g,%.17g" % (
            query.bbox[1],
            query.bbox[0],
            query.bbox[3],
            query.bbox[2],
        )
    else:
        headers["Content-Bbox"] = "%.17g,%.17g,%.17g,%.17g" % (
            query.bbox[0],
            query.bbox[1],
            query.bbox[2],
            query.bbox[3],
        )
    return headers


def get_width_height(server: OGCAPIServer, req: Request):
    width = req.args.get("width", None)
    height = req.args.get("height", None)

    if width:
        try:
            width = int(width)
        except ValueError:
            raise OGCAPIServer.invalid_parameter("width is not an integer")
        if width <= 0:
            raise OGCAPIServer.invalid_parameter("width must be strictly positive")

    if height:
        try:
            height = int(height)
        except ValueError:
            raise OGCAPIServer.invalid_parameter("height is not an integer")
        if height <= 0:
            raise OGCAPIServer.invalid_parameter("height must be strictly positive")

    return width, height


def check_width_height(server: OGCAPIServer, req: Request, width: int, height: int):
    if server.max_width and width > server.max_width:
        raise OGCAPIServer.invalid_parameter(
            f"width={width} exceeds max_width={server.max_width}"
        )
    if server.max_height and height > server.max_height:
        raise OGCAPIServer.invalid_parameter(
            f"height={height} exceeds max_height={server.max_height}"
        )
    if server.max_output_pixels and width * height > server.max_output_pixels:
        raise OGCAPIServer.invalid_parameter(
            f"number of pixels {width * height} exceed max_pixels={server.max_output_pixels}"
        )


def get_map_query(
    server: OGCAPIServer, req: Request, layer: WMSLayer, format: str
) -> (MapQuery, ImageOptions):
    """Return a tuple (MapQuery, ImageOptions) from the incoming request"""
    log = server.log

    bbox_crs = req.args.get("bbox-crs", None)
    bbox = req.args.get("bbox", None)
    subset_crs = req.args.get("subset-crs", None)
    subset = (
        req.args.get_all("subset")
        if not isinstance(req.args, dict)
        else req.args.get("subset", None)
    )  # potentially multi valued
    center_crs = req.args.get("center-crs", None)
    center = req.args.get("center", None)
    width, height = get_width_height(server, req)
    scale_denominator = req.args.get("scale-denominator", None)

    try:
        mm_per_pixel = float(req.args.get("mm-per-pixel", "0.28"))
        if mm_per_pixel <= 0:
            raise ValueError
    except ValueError:
        raise OGCAPIServer.invalid_parameter(
            "mm-per-pixel must be a strictly positive numeric value"
        )
    display_res_m_per_px = mm_per_pixel / 1000.0

    if (bbox or center) and subset:
        raise OGCAPIServer.invalid_parameter(
            "(bbox or center) and subset are mutually exclusive"
        )

    crs = req.args.get("crs", None)
    if crs:
        try:
            crs = ogcapi_srs_to_srs(crs)
        except ValueError:
            raise OGCAPIServer.invalid_parameter("crs is not a valid CRS")

        if crs not in server.map_srs:
            raise OGCAPIServer.invalid_parameter("crs is incompatible with this layer")
    elif len(server.map_srs) > 0:
        crs = server.map_srs[0]
    else:
        crs = SRS(4326)

    if subset:
        subset_crs = get_bbox_center_or_subset_crs(subset_crs, layer, "subset-crs")

        bbox = subset_to_bbox(subset)
        if bbox[2] <= bbox[0] or bbox[3] <= bbox[1]:
            raise OGCAPIServer.invalid_parameter(
                "Invalid range in subset: first value must be less than second one"
            )

        try:
            bbox = MapExtent(bbox, subset_crs).transform(crs).bbox
        except Exception as e:
            raise OGCAPIServer.invalid_parameter(
                f"subset cannot be reprojected from {subset_crs} to {crs}: {e}"
            )

    elif bbox:
        bbox_crs = get_bbox_center_or_subset_crs(bbox_crs, layer, "bbox-crs")
        try:
            bbox = [float(x) for x in bbox.split(",")]
            if len(bbox) != 4 and len(bbox) != 6:
                raise ValueError
        except ValueError:  # catch ValueError from float or local raise
            raise OGCAPIServer.invalid_parameter(
                "bbox must be a list of 4 or 6 numeric values"
            )
        if len(bbox) == 6:
            # Ignore minz / maxz
            bbox = [bbox[0], bbox[1], bbox[3], bbox[4]]
        if bbox[2] <= bbox[0] or bbox[3] <= bbox[1]:
            raise OGCAPIServer.invalid_parameter(
                "bbox[2] must be greater than bbox[0] and bbox[3] must be greater than bbox[1]"
            )

        if bbox_crs.is_axis_order_ne:
            bbox = [bbox[1], bbox[0], bbox[3], bbox[2]]

        try:
            bbox = MapExtent(bbox, bbox_crs).transform(crs).bbox
        except Exception as e:
            raise OGCAPIServer.invalid_parameter(
                f"bbox cannot be reprojected from {bbox_crs} to {crs}: {e}"
            )

    elif center:
        center_crs = get_bbox_center_or_subset_crs(center_crs, layer, "center-crs")

        try:
            center = [float(x) for x in center.split(",")]
            if len(center) != 2:
                raise ValueError
        except ValueError:  # catch ValueError from float or local raise
            raise OGCAPIServer.invalid_parameter(
                "center must be a list of 2 numeric values"
            )

        if center_crs.is_axis_order_ne:
            new_center = [center[1], center[0]]
            center = new_center

        try:
            center = list(center_crs.transform_to(crs, (center[0], center[1])))
        except Exception as e:
            raise OGCAPIServer.invalid_parameter(
                f"center cannot be reprojected from {center_crs} to {crs}: {e}"
            )

    if scale_denominator:
        try:
            scale_denominator = float(scale_denominator)
        except ValueError:
            raise OGCAPIServer.invalid_parameter(
                "scale-denominator must be a strictly positive numeric value"
            )
        if scale_denominator <= 0:
            raise OGCAPIServer.invalid_parameter(
                "scale-denominator must be a strictly positive numeric value"
            )

    # Cf https://docs.ogc.org/is/20-058/20-058.html#overview-subsetting-and-scaling

    # Cf Table 5 "Always valid requests (no scaling or subsetting parameter)"
    if not (bbox or center or width or height or scale_denominator):
        bbox = layer.extent.transform(crs).bbox
        aspect_ratio = get_aspect_ratio(bbox)
        width, height = get_width_height_from_default_size(aspect_ratio)

    # Cf Table 6 "Always invalid parameter combinations"
    elif bbox and center:
        raise OGCAPIServer.invalid_parameter("bbox and center are mutually exclusive")
    elif bbox and scale_denominator and (width or height):
        raise OGCAPIServer.invalid_parameter(
            "bbox and scale-denominator and (width/height) are mutually exclusive"
        )

    # For all cases below, cf Table 9
    # "Parameter combinations for implementations supporting both Subsetting and Scaling"
    else:
        width, height, bbox = compute_width_height_bbox(
            server,
            req,
            width,
            height,
            bbox,
            crs,
            center,
            scale_denominator,
            display_res_m_per_px,
            layer,
        )

    log.debug(f"width={width}, height={height}, bbox={bbox}, crs={crs}")
    assert width
    assert height
    assert bbox

    check_width_height(server, req, width, height)

    query = MapQuery(bbox, (width, height), crs, format=format)

    image_opts = server.image_formats[FORMAT_TYPES[format]].copy()
    bgcolor = get_bgcolor(server, req)
    if bgcolor:
        image_opts.bgcolor = bgcolor
    image_opts.transparent = get_transparent(req, bgcolor)

    return query, image_opts


def get_bgcolor(server: OGCAPIServer, req: Request):
    bgcolor = req.args.get("bgcolor", None)
    if bgcolor:
        bgcolor = bgcolor.lower()
        if bgcolor.startswith("0x") and len(bgcolor) == 8:
            try:
                R = int(bgcolor[2:4], base=16)
                G = int(bgcolor[4:6], base=16)
                B = int(bgcolor[6:8], base=16)
                bgcolor = (R, G, B)
            except ValueError:
                raise OGCAPIServer.invalid_parameter("invalid value for bgcolor")
        elif bgcolor.startswith("0x") and len(bgcolor) == 10:
            try:
                A = int(bgcolor[2:4], base=16)
                R = int(bgcolor[4:6], base=16)
                G = int(bgcolor[6:8], base=16)
                B = int(bgcolor[8:10], base=16)
                bgcolor = (R, G, B, A)
            except ValueError:
                raise OGCAPIServer.invalid_parameter("invalid value for bgcolor")
        else:
            if bgcolor in W3C_BGCOLORS:
                bgcolor = W3C_BGCOLORS[bgcolor]
                assert len(bgcolor) == 7 and bgcolor[0] == "#"
                bgcolor = (
                    int(bgcolor[1:3], base=16),
                    int(bgcolor[3:5], base=16),
                    int(bgcolor[5:7], base=16),
                )
            else:
                raise OGCAPIServer.invalid_parameter("invalid value for bgcolor")

    return bgcolor


def get_transparent(req: Request, bgcolor):
    transparent = req.args.get("transparent", None)
    if transparent is not None:
        if transparent == "true":
            transparent = True
        elif transparent == "false":
            transparent = False
        else:
            raise OGCAPIServer.invalid_parameter("invalid value for transparent")
    elif bgcolor:
        transparent = True

    return transparent


def render_map(
    server: OGCAPIServer,
    req: Request,
    query: MapQuery,
    image_opts: ImageOptions,
    layer: WMSLayer,
) -> Response:
    """Return the image in a Response object from the query."""

    # limit query to srs_extent if query is larger
    orig_query = query
    query_extent = MapExtent(query.bbox, query.srs)
    if not layer.extent.transform(query.srs).contains(query_extent):
        limited_extent = layer.extent.transform(query.srs).intersection(query_extent)
        if not limited_extent:
            img = BlankImageSource(
                size=query.size, image_opts=image_opts, cacheable=True
            )
            return Response(
                img.as_buffer(),
                content_type=FORMAT_TYPES[query.format],
                headers=get_headers(server, query),
                status=200,
            )

        sub_size, offset, sub_bbox = bbox_position_in_image(
            query.bbox, query.size, limited_extent.bbox
        )
        query = MapQuery(sub_bbox, sub_size, query.srs, query.format)

    actual_layers = OrderedDict()
    # only add if layer renders the query
    if layer.renders_query(query):
        # if layer is not transparent and will be rendered,
        # remove already added (then hidden) layers
        if layer.is_opaque(query):
            actual_layers = OrderedDict()
        for layer_name, map_layers in layer.map_layers_for_query(query):
            actual_layers[layer_name] = map_layers

    render_layers = []
    for layers in actual_layers.values():
        render_layers.extend(layers)

    raise_source_errors = True if server.on_error == "raise" else False
    renderer = LayerRenderer(
        render_layers,
        query,
        req,
        raise_source_errors=raise_source_errors,
        concurrent_rendering=server.concurrent_layer_renderer,
    )

    merger = LayerMerger()
    renderer.render(merger)

    if server.attribution and server.attribution.get("text"):
        merger.add(attribution_image(server.attribution["text"], query.size))

    try:
        result = merger.merge(
            size=query.size, image_opts=image_opts, bbox=query.bbox, bbox_srs=query.srs
        )
    except Exception as ex:
        raise OGCAPIServer.exception(
            "InternalError",
            "error while merging layers: %s" % str(ex),
            status=500,
        )

    if query != orig_query:
        result = SubImageSource(
            result, size=orig_query.size, offset=offset, image_opts=image_opts
        )

    # Provide the wrapping WSGI app or filter the opportunity to process the
    # image before it's wrapped up in a response
    result = server.decorate_img(
        result,
        "ogcapi.map",
        list(actual_layers.keys()),
        req.environ,
        (query.srs.srs_code, query.bbox),
    )

    query = orig_query

    try:
        result.georef = GeoReference(bbox=query.bbox, srs=query.srs)
        result_buf = result.as_buffer(image_opts)
    except IOError as ex:
        raise OGCAPIServer.exception(
            "InternalError",
            "error while processing image file: %s" % ex,
            status=500,
        )

    return Response(
        result_buf,
        content_type=FORMAT_TYPES[query.format],
        headers=get_headers(server, query),
        status=200,
    )


def get_map(
    server: OGCAPIServer, req: Request, coll_id: Optional[str], map_filename: str
):
    log = server.log
    log.debug(f"Map for {coll_id}")

    allowed_args = set(
        [
            "f",
            "crs",
            "bbox-crs",
            "bbox",
            "center-crs",
            "center",
            "subset",
            "subset-crs",
            "width",
            "height",
            "scale-denominator",
            "mm-per-pixel",
            "bgcolor",
            "transparent",
        ]
    )
    collections = req.args.get("collections", None)

    if coll_id is None and (server.default_dataset_layers or collections):
        allowed_args.add("collections")

    if req.args.get("service", None) == "WMS":
        # Hack for the OpenLayers map in our collection.html that issues WMS requests...
        assert "width" in req.args
        assert "height" in req.args
        assert "bbox" in req.args
        assert req.args["version"] == "1.1.1"
        srs = SRS(req.args["srs"])
        if srs.is_axis_order_ne:
            bbox = req.args["bbox"].split(",")
            bbox = "%s,%s,%s,%s" % (bbox[1], bbox[0], bbox[3], bbox[2])
            req.args["bbox"] = bbox
        req.args["bbox-crs"] = req.args["srs"]
        req.args["crs"] = req.args["srs"]
    else:
        for arg in req.args:
            if arg not in allowed_args:
                raise OGCAPIServer.unknown_query_parameter(arg)
        if req.args.get("f", None) not in (None, F_PNG, F_JPEG):
            raise OGCAPIServer.invalid_parameter("Invalid value for f query parameter")

    if coll_id is None:
        # Dataset map
        if collections is None and server.default_dataset_layers is None:
            raise OGCAPIServer.exception(
                "InternalError",
                "Dataset map request but no default_dataset_layers defined in server configuration",
                status=500,
            )

        if collections:
            layers = []
            for col in collections.split(","):
                if col not in server.layers:
                    raise OGCAPIServer.invalid_parameter(f"unknown collection {col}")
                layers.append(server.layers[col])
        else:
            layers = server.default_dataset_layers
        layer = WMSGroupLayer("group_layer", "group_layer", None, layers)
    else:
        if coll_id not in server.layers:
            raise OGCAPIServer.collection_not_found()
        layer = server.layers[coll_id]

    format = get_image_format(req, map_filename)

    query, image_opts = get_map_query(server, req, layer, format)
    return render_map(server, req, query, image_opts, layer)



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division

import os
import tempfile

from io import BytesIO
from contextlib import contextmanager

from PIL import Image, ImageDraw, ImageColor


def assert_image_mode(img, mode):
    pos = img.tell()
    try:
        img = Image.open(img)
        assert img.mode == mode
    finally:
        img.seek(pos)


def check_format(img, format):
    assert globals()['is_' + format.lower()](img), 'img is not %s' % format


def has_magic_bytes(fileobj, bytes):
    pos = fileobj.tell()
    for magic in bytes:
        fileobj.seek(0)
        it_is = fileobj.read(len(magic)) == magic
        fileobj.seek(pos)
        if it_is:
            return True
    return False


magic_bytes = {'png': [b"\211PNG\r\n\032\n"],
               'tiff': [b"MM\x00\x2a", b"II\x2a\x00"],
               'geotiff': [b"MM\x00\x2a", b"II\x2a\x00"],
               'gif': [b"GIF87a", b"GIF89a"],
               'jpeg': [b"\xFF\xD8"],
               'bmp': [b'BM']
               }


def create_is_x_functions():
    for type_, magic in magic_bytes.items():
        def create_is_type(type_, magic):
            def is_type(fileobj):
                if not hasattr(fileobj, 'read'):
                    fileobj = BytesIO(fileobj)
                return has_magic_bytes(fileobj, magic)
            return is_type
        globals()['is_' + type_] = create_is_type(type_, magic)


create_is_x_functions()
del create_is_x_functions


def is_transparent(img_data):
    data = BytesIO(img_data)
    img = Image.open(data)
    if img.mode == 'P':
        img = img.convert('RGBA')
    if img.mode == 'RGBA':
        return any(img.histogram()[-256:-1])

    raise NotImplementedError(
        'assert_is_transparent works only for RGBA images, got %s image' % img.mode)


def img_from_buf(buf):
    data = BytesIO(buf)
    return Image.open(data)


def bgcolor_ratio(img_data):
    """
    Return the ratio of the primary/bg color. 1 == only bg color.
    """
    data = BytesIO(img_data)
    img = Image.open(data)
    total_colors = img.size[0] * img.size[1]
    colors = img.getcolors()
    colors.sort()
    bgcolor = colors[-1][0]
    return bgcolor/total_colors


def create_tmp_image_file(size, two_colored=False):
    fd, out_file = tempfile.mkstemp(suffix='.png')
    os.close(fd)
    img = Image.new('RGBA', size)
    if two_colored:
        draw = ImageDraw.Draw(img)
        draw.rectangle((0, 0, img.size[0]//2, img.size[1]),
                       fill=ImageColor.getrgb('white'))
    img.save(out_file, 'png')
    return out_file


def create_image(size, color=None, mode=None):
    if color is not None:
        if isinstance(color, str):
            if mode is None:
                mode = 'RGB'
            img = Image.new(mode, size, color=color)
        else:
            if mode is None:
                mode = 'RGBA' if len(color) == 4 else 'RGB'
            img = Image.new(mode, size, color=tuple(color))
    else:
        img = create_debug_img(size)
    return img


def create_tmp_image_buf(size, format='png', color=None, mode='RGB'):
    img = create_image(size, color, mode)
    data = BytesIO()
    img.save(data, format)
    data.seek(0)
    return data


def create_tmp_image(size, format='png', color=None, mode='RGB'):
    data = create_tmp_image_buf(size, format, color, mode)
    return data.read()


def create_debug_img(size, transparent=True):
    if transparent:
        img = Image.new("RGBA", size)
    else:
        img = Image.new("RGB", size, ImageColor.getrgb("#EEE"))

    draw = ImageDraw.Draw(img)
    draw_pattern(draw, size)
    return img


def draw_pattern(draw, size):
    w, h = size
    black_color = ImageColor.getrgb("black")
    draw.rectangle((0, 0, w-1, h-1), outline=black_color)
    draw.ellipse((0, 0, w-1, h-1), outline=black_color)
    step = w/16.0
    for i in range(16):
        color = ImageColor.getrgb('#3' + hex(16-i)[-1] + hex(i)[-1])
        draw.line((i*step, 0, i*step, h), fill=color)
    step = h/16.0
    for i in range(16):
        color = ImageColor.getrgb('#' + hex(16-i)[-1] + hex(i)[-1] + '3')
        draw.line((0, i*step, w, i*step), fill=color)


@contextmanager
def tmp_image(size, format='png', color=None, mode='RGB'):
    if color is not None:
        img = Image.new(mode, size, color=color)
    else:
        img = create_debug_img(size)
    if format == 'jpeg':
        img = img.convert('RGB')
    data = BytesIO()
    img.save(data, format)
    data.seek(0)
    yield data


def assert_img_colors_eq(img1, img2, delta=1, pixel_delta=1):
    """
    assert that the colors of two images are equal.
    Use `delta` to accept small color variations
    (e.g. (255, 0, 127) == (254, 1, 128) with delta=1)
    Use `pixel_delta` to accept small variations in the number of pixels for each color
    (in percent of total pixels).

    `img1` and `img2` needs to be an image or list of
    colors like ``[(n, (r, g, b)), (n, (r, g, b)), ...]``
    """
    colors1 = sorted(img1.getcolors() if hasattr(img1, 'getcolors') else img1)
    colors2 = sorted(img2.getcolors() if hasattr(img2, 'getcolors') else img2)

    total_pixels = sum(n for n, _ in colors1)
    for (n1, c1), (n2, c2) in zip(colors1, colors2):
        assert abs(n1 - n2) < (total_pixels / 100 * pixel_delta), 'num colors not equal: %r != %r' % (colors1, colors2)
        assert_colors_eq(c1, c2)


assert_colors_equal = assert_img_colors_eq


def assert_colors_eq(c1, c2, delta=1):
    """
    assert that two colors are equal. Use `delta` to accept
    small color variations.
    """
    assert abs(c1[0] - c2[0]) <= delta, 'colors not equal: %r != %r' % (c1, c2)
    assert abs(c1[1] - c2[1]) <= delta, 'colors not equal: %r != %r' % (c1, c2)
    assert abs(c1[2] - c2[2]) <= delta, 'colors not equal: %r != %r' % (c1, c2)



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import logging
import os
from PIL import Image, ImageColor, ImageDraw, ImageFont
try:
    import importlib_resources
except ImportError:
    from importlib import resources as importlib_resources

from mapproxy.config import base_config, abspath
from mapproxy.image import ImageSource
from mapproxy.image.opts import create_image, ImageOptions

_pil_ttf_support = True


log_system = logging.getLogger('mapproxy.system')


def message_image(message, size, image_opts, bgcolor='#ffffff',
                  transparent=False):
    """
    Creates an image with text (`message`). This can be used
    to create in_image exceptions.

    For dark `bgcolor` the font color is white, otherwise black.

    :param message: the message to put in the image
    :param size: the size of the output image
    :param format: the output format of the image
    :param bgcolor: the background color of the image
    :param transparent: if True and the `format` supports it,
                        return a transparent image
    :rtype: `ImageSource`
    """
    eimg = ExceptionImage(message, image_opts=image_opts)
    return eimg.draw(size=size)


def attribution_image(message, size, image_opts=None, inverse=False):
    """
    Creates an image with text attribution (`message`).

    :param message: the message to put in the image
    :param size: the size of the output image
    :param format: the output format of the image
    :param inverse: if true, write white text
    :param transparent: if True and the `format` supports it,
                        return a transparent image
    :rtype: `ImageSource`
    """
    if image_opts is None:
        image_opts = ImageOptions(transparent=True)
    aimg = AttributionImage(message, image_opts=image_opts,
                            inverse=inverse)
    return aimg.draw(size=size)


class MessageImage(object):
    """
    Base class for text rendering in images (for watermarks, exception images, etc.)

    :ivar font_name: the font name for the text
    :ivar font_size: the font size of the text
    :ivar font_color: the color of the font as a tuple
    :ivar box_color: the color of the box behind the text.
                     color as a tuple or ``None``
    """
    font_name = 'DejaVu Sans Mono'
    font_size = 10
    font_color = ImageColor.getrgb('black')
    box_color = None
    linespacing = 5
    padding = 3
    placement = 'ul'

    def __init__(self, message, image_opts):
        self.message = message
        self.image_opts = image_opts
        self._font = None

    @property
    def font(self):
        global _pil_ttf_support
        if self._font is None:
            if self.font_name != 'default' and _pil_ttf_support:
                try:
                    self._font = ImageFont.truetype(font_file(self.font_name),
                                                    self.font_size)
                except ImportError:
                    _pil_ttf_support = False
                    log_system.warning("Couldn't load TrueType fonts, "
                                       "PIL needs to be build with freetype support.")
                except IOError:
                    _pil_ttf_support = False
                    log_system.warning("Couldn't load find TrueType font ", self.font_name)
            if self._font is None:
                self._font = ImageFont.load_default()
        return self._font

    def new_image(self, size):
        return Image.new('RGBA', size)

    def draw(self, img=None, size=None, in_place=True):
        """
        Create the message image. Either draws on top of `img` or creates a
        new image with the given `size`.
        """
        if not ((img and not size) or (size and not img)):
            raise TypeError('need either img or size argument')

        if img is None:
            base_img = self.new_image(size)
        elif not in_place:
            size = img.size
            base_img = self.new_image(size)
        else:
            base_img = img.as_image()
            size = base_img.size

        if not self.message:
            if img is not None:
                return img
            return ImageSource(base_img, size=size, image_opts=self.image_opts)

        draw = ImageDraw.Draw(base_img)
        self.draw_msg(draw, size)
        image_opts = self.image_opts
        if not in_place and img:
            image_opts = image_opts or img.image_opts
            img = img.as_image()
            converted = False
            if len(self.font_color) == 4 and img.mode != 'RGBA':
                # we need RGBA to keep transparency from text
                converted = img.mode
                img = img.convert('RGBA')
            img.paste(base_img, (0, 0), base_img)
            if converted == 'RGB':
                # convert image back
                img = img.convert('RGB')
            base_img = img

        return ImageSource(base_img, size=size, image_opts=image_opts)

    def draw_msg(self, draw, size):
        td = TextDraw(self.message, font=self.font, bg_color=self.box_color,
                      font_color=self.font_color, placement=self.placement,
                      linespacing=self.linespacing, padding=self.padding)
        td.draw(draw, size)


class ExceptionImage(MessageImage):
    """
    Image for exceptions.
    """
    font_name = 'default'
    font_size = 9

    def __init__(self, message, image_opts):
        MessageImage.__init__(self, message, image_opts=image_opts.copy())
        if not self.image_opts.bgcolor:
            self.image_opts.bgcolor = '#ffffff'

    def new_image(self, size):
        return create_image(size, self.image_opts)

    @property
    def font_color(self):
        if self.image_opts.transparent:
            return ImageColor.getrgb('black')
        if _luminance(ImageColor.getrgb(self.image_opts.bgcolor)) < 128:
            return ImageColor.getrgb('white')
        return ImageColor.getrgb('black')


class WatermarkImage(MessageImage):
    """
    Image with large, faded message.
    """
    font_name = 'DejaVu Sans'
    font_size = 24
    font_color = (128, 128, 128)

    def __init__(self, message, image_opts, placement='c', opacity=None, font_color=None, font_size=None):
        MessageImage.__init__(self, message, image_opts=image_opts)
        if opacity is None:
            opacity = 30
        if font_size:
            self.font_size = font_size
        if font_color:
            self.font_color = font_color
        self.font_color = self.font_color + tuple([opacity])
        self.placement = placement

    def draw_msg(self, draw, size):
        td = TextDraw(self.message, self.font, self.font_color)
        if self.placement in ('l', 'b'):
            td.placement = 'cL'
            td.draw(draw, size)
        if self.placement in ('r', 'b'):
            td.placement = 'cR'
            td.draw(draw, size)
        if self.placement == 'c':
            td.placement = 'cc'
            td.draw(draw, size)


class AttributionImage(MessageImage):
    """
    Image with attribution information.
    """
    font_name = 'DejaVu Sans'
    font_size = 10
    placement = 'lr'

    def __init__(self, message, image_opts, inverse=False):
        MessageImage.__init__(self, message, image_opts=image_opts)
        self.inverse = inverse

    @property
    def font_color(self):
        if self.inverse:
            return ImageColor.getrgb('white')
        else:
            return ImageColor.getrgb('black')

    @property
    def box_color(self):
        if self.inverse:
            return (0, 0, 0, 100)
        else:
            return (255, 255, 255, 120)


class TextDraw(object):
    def __init__(self, text, font, font_color=None, bg_color=None,
                 placement='ul', padding=5, linespacing=3):
        if isinstance(text, str):
            text = text.split('\n')
        self.text = text
        self.font = font
        self.bg_color = bg_color
        self.font_color = font_color
        self.placement = placement
        self.padding = (padding, padding, padding, padding)
        self.linespacing = linespacing

    def text_boxes(self, draw, size):
        try:
            total_bbox, boxes = self._relative_text_boxes(draw)
        except UnicodeEncodeError:
            # raised if font does not support unicode
            self.text = [x.encode('ascii', 'replace') for x in self.text]
            total_bbox, boxes = self._relative_text_boxes(draw)
        return self._place_boxes(total_bbox, boxes, size)

    def draw(self, draw, size):
        total_bbox, boxes = self.text_boxes(draw, size)
        if self.bg_color:
            draw.rectangle(
                (total_bbox[0]-self.padding[0],
                 total_bbox[1]-self.padding[1],
                 total_bbox[2]+self.padding[2],
                 total_bbox[3]+self.padding[3]),
                fill=self.bg_color)

        for text, box in zip(self.text, boxes):
            draw.text((box[0], box[1]), text, font=self.font, fill=self.font_color)

    def _relative_text_boxes(self, draw):
        total_bbox = (1e9, 1e9, -1e9, -1e9)
        boxes = []
        y_offset = 0
        for i, line in enumerate(self.text):
            try:
                text_box = draw.textbbox((0, y_offset), line, font=self.font)
                y_offset = text_box[3] + self.linespacing
            except AttributeError:
                # Some Pillow versions do not support textbox for non true type fonts
                text_size = draw.textsize(line, font=self.font)
                text_box = (0, y_offset, text_size[0], text_size[1]+y_offset)
                y_offset += text_size[1] + self.linespacing
            boxes.append(text_box)
            total_bbox = (min(total_bbox[0], text_box[0]),
                          min(total_bbox[1], text_box[1]),
                          max(total_bbox[2], text_box[2]),
                          max(total_bbox[3], text_box[3]),
                          )

        return total_bbox, boxes

    def _move_bboxes(self, boxes, offsets):
        result = []
        for box in boxes:
            box = box[0]+offsets[0], box[1]+offsets[1], box[2]+offsets[0], box[3]+offsets[1]
            result.append(tuple(int(x) for x in box))
        return result

    def _place_boxes(self, total_bbox, boxes, size):
        x_offset = y_offset = None
        text_size = (total_bbox[2] - total_bbox[0]), (total_bbox[3] - total_bbox[1])

        if self.placement[0] == 'u':
            y_offset = self.padding[1]
        elif self.placement[0] == 'l':
            y_offset = size[1] - self.padding[3] - text_size[1]
        elif self.placement[0] == 'c':
            y_offset = size[1] // 2 - text_size[1] // 2

        if self.placement[1] == 'l':
            x_offset = self.padding[0]
        if self.placement[1] == 'L':
            x_offset = -text_size[0] // 2
        elif self.placement[1] == 'r':
            x_offset = size[0] - self.padding[1] - text_size[0]
        elif self.placement[1] == 'R':
            x_offset = size[0] - text_size[0] // 2
        elif self.placement[1] == 'c':
            x_offset = size[0] // 2 - text_size[0] // 2

        if x_offset is None or y_offset is None:
            raise ValueError('placement %r not supported' % self.placement)

        offsets = x_offset, y_offset
        return self._move_bboxes([total_bbox], offsets)[0], self._move_bboxes(boxes, offsets)


def font_file(font_name):
    font_dir = base_config().image.font_dir
    font_name = font_name.replace(' ', '')
    if font_dir:
        abspath(font_dir)
        path = os.path.join(font_dir, font_name + '.ttf')
    else:
        path = str(importlib_resources.files(__package__).joinpath('fonts').joinpath(font_name + '.ttf'))
    return path


def _luminance(color):
    """
    Returns the luminance of a RGB tuple. Uses ITU-R 601-2 luma transform.
    """
    r, g, b = color
    return r * 299/1000 + g * 587/1000 + b * 114/1000



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from mapproxy.config import Options, base_config, load_base_config
from mapproxy.test.helper import TempFiles


class TestOptions(object):
    def test_update_overwrite(self):
        d = Options(foo='bar', baz=4)
        d.update(Options(baz=5))
        assert d.baz == 5
        assert d.foo == 'bar'

    def test_update_new(self):
        d = Options(foo='bar', baz=4)
        d.update(Options(biz=5))
        assert d.baz == 4
        assert d.biz == 5
        assert d.foo == 'bar'

    def test_update_recursive(self):
        d = Options(
            foo='bar',
            baz=Options(ham=2, eggs=4))
        d.update(Options(baz=Options(eggs=5)))
        assert d.foo == 'bar'
        assert d.baz.ham == 2
        assert d.baz.eggs == 5

    def test_compare(self):
        assert Options(foo=4) == Options(foo=4)
        assert Options(foo=Options(bar=4)) == Options(foo=Options(bar=4))


class TestDefaultsLoading(object):
    def teardown_method(self):
        load_base_config(clear_existing=True)

    defaults_yaml = b"""
    foo:
        bar:
            ham: 2
            eggs: 4
    biz: 'foobar'
    wiz: 'foobar'
    """

    def test_defaults(self):
        with TempFiles() as tmp:
            with open(tmp[0], 'wb') as f:
                f.write(TestDefaultsLoading.defaults_yaml)
            load_base_config(config_file=tmp[0], clear_existing=True)

            assert base_config().biz == 'foobar'
            assert base_config().wiz == 'foobar'
            assert base_config().foo.bar.ham == 2
            assert base_config().foo.bar.eggs == 4
            assert not hasattr(base_config(), 'wms')

    def test_defaults_overwrite(self):
        with TempFiles(2) as tmp:
            with open(tmp[0], 'wb') as f:
                f.write(TestDefaultsLoading.defaults_yaml)
            with open(tmp[1], 'wb') as f:
                f.write(b"""
                baz: [9, 2, 1, 4]
                biz: 'barfoo'
                foo:
                    bar:
                        eggs: 5
                """)

            load_base_config(config_file=tmp[0], clear_existing=True)
            load_base_config(config_file=tmp[1])

            assert base_config().biz == 'barfoo'
            assert base_config().wiz == 'foobar'
            assert base_config().baz == [9, 2, 1, 4]
            assert base_config().foo.bar.ham == 2
            assert base_config().foo.bar.eggs == 5
            assert not hasattr(base_config(), 'wms')


class TestSRSConfig(object):
    def teardown_method(self):
        load_base_config(clear_existing=True)

    def test_user_srs_definitions(self):
        user_yaml = b"""
        srs:
          axis_order_ne: ['EPSG:9999']
        """
        with TempFiles() as tmp:
            with open(tmp[0], 'wb') as f:
                f.write(user_yaml)

            load_base_config(config_file=tmp[0])

            assert 'EPSG:9999' in base_config().srs.axis_order_ne
            assert 'EPSG:9999' not in base_config().srs.axis_order_en

            # defaults still there
            assert 'EPSG:31468' in base_config().srs.axis_order_ne
            assert 'CRS:84' in base_config().srs.axis_order_en



# This file is part of the MapProxy project.
# Copyright (C) 2017 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import hashlib
import datetime
import time
from io import BytesIO

from mapproxy.image import ImageSource
from mapproxy.cache.base import (
    TileCacheBase,
    tile_buffer,
)

try:
    import redis
except ImportError:
    redis = None


import logging
log = logging.getLogger(__name__)


class RedisCache(TileCacheBase):
    def __init__(
            self, host, port, prefix, ttl=0, db=0, username=None, password=None, coverage=None, ssl_certfile=None,
            ssl_keyfile=None, ssl_ca_certs=None):
        super(RedisCache, self).__init__(coverage)

        if redis is None:
            raise ImportError("Redis backend requires 'redis' package.")

        self.ssl_certfile = ssl_certfile
        self.ssl_keyfile = ssl_keyfile
        self.ssl_ca_certs = ssl_ca_certs

        self.prefix = prefix
        # str(username) => if not set defaults to None
        md5 = hashlib.new('md5', (host + str(port) + prefix + str(db)).encode('utf-8'), usedforsecurity=False)
        self.lock_cache_id = 'redis-' + md5.hexdigest()
        self.ttl = ttl
        # Enable SSL only if certificate and key are provided (CA certificates are not mandatory, but if provided use
        # them)
        ssl_enabled = all([self.ssl_certfile, self.ssl_keyfile])
        ssl_certfile = self.ssl_certfile if ssl_enabled else None
        ssl_keyfile = self.ssl_keyfile if ssl_enabled else None
        ssl_ca_certs = self.ssl_ca_certs if ssl_enabled and self.ssl_ca_certs else None
        self.r = redis.StrictRedis(
            host=host,
            port=port,
            username=username,
            password=password,
            db=db,
            ssl_certfile=ssl_certfile,
            ssl_keyfile=ssl_keyfile,
            ssl_ca_certs=ssl_ca_certs,
            ssl=ssl_enabled
        )

    def _key(self, tile):
        x, y, z = tile.coord
        return self.prefix + '-%d-%d-%d' % (z, x, y)

    def is_cached(self, tile, dimensions=None):
        if tile.coord is None or tile.source:
            return True
        key = self._key(tile)

        try:
            log.debug('exists_key, key: %s' % key)
            return self.r.exists(key)
        except redis.exceptions.ConnectionError as e:
            log.error('Error during connection %s' % e)
            return False
        except Exception as e:
            log.error('REDIS:exists_key error  %s' % e)
            return False

    def store_tile(self, tile, dimensions=None):
        if tile.stored:
            return True
        key = self._key(tile)

        with tile_buffer(tile) as buf:
            data = buf.read()

        try:
            log.debug('store_key, key: %s' % key)
            r = self.r.set(key, data)
        except redis.exceptions.ConnectionError as e:
            log.error('Error during connection %s' % e)
            return False
        except Exception as e:
            log.error('REDIS:store_key error  %s' % e)
            return False

        if self.ttl:
            # use ms expire times for unit-tests
            self.r.pexpire(key, int(self.ttl * 1000))
        return r

    def load_tile_metadata(self, tile, dimensions=None):
        if tile.timestamp:
            return
        pipe = self.r.pipeline()
        pipe.ttl(self._key(tile))
        pipe.memory_usage(self._key(tile))
        pipe_res = pipe.execute()
        tile.timestamp = time.mktime(datetime.datetime.now().timetuple()) - self.ttl - int(pipe_res[0])
        tile.size = pipe_res[1]

    def load_tile(self, tile, with_metadata=False, dimensions=None):
        if tile.source or tile.coord is None:
            return True
        key = self._key(tile)

        try:
            log.debug('get_key, key: %s' % key)
            tile_data = self.r.get(key)
            if tile_data:
                tile.source = ImageSource(BytesIO(tile_data))
                return True
            return False
        except redis.exceptions.ConnectionError as e:
            log.error('Error during connection %s' % e)
            return False
        except Exception as e:
            log.error('REDIS:get_key error  %s' % e)
            return False

    def remove_tile(self, tile, dimensions=None):
        if tile.coord is None:
            return True

        key = self._key(tile)
        self.r.delete(key)
        return True



def escape_html(data):
    """
    Escape user-provided input data for safe inclusion in HTML _and_ JS to prevent XSS.
    """
    data = data.replace('&', '&amp;')
    data = data.replace('>', '&gt;')
    data = data.replace('<', '&lt;')
    data = data.replace("'", '')
    data = data.replace('"', '')
    return data



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tile caching (creation, caching and retrieval of tiles).

.. digraph:: Schematic Call Graph

    ranksep = 0.1;
    node [shape="box", height="0", width="0"]

    cl  [label="CacheMapLayer" href="<mapproxy.core.layer.CacheMapLayer>"]
    tm  [label="TileManager",  href="<TileManager>"];
    fc      [label="FileCache", href="<FileCache>"];
    s       [label="Source", href="<mapproxy.core.source.Source>"];

    {
        cl -> tm [label="load_tile_coords"];
        tm -> fc [label="load\\nstore\\nis_cached"];
        tm -> s  [label="get_map"]
    }


"""
from functools import partial
from contextlib import contextmanager
from mapproxy.grid.meta_grid import MetaGrid
from mapproxy.image import BlankImageSource
from mapproxy.image.mask import mask_image_source_from_coverage
from mapproxy.image.opts import ImageOptions
from mapproxy.image.merge import merge_images
from mapproxy.image.tile import TileSplitter, TiledImage
from mapproxy.layer import MapQuery, BlankImage
from mapproxy.source import SourceError, DummySource
from mapproxy.util import async_
from mapproxy.util.py import reraise, reraise_exception
import sys
import logging
log = logging.getLogger('mapproxy.cache.tile')


class TileManager(object):
    """
    Manages tiles for a single grid.
    Loads tiles from the cache, creates new tiles from sources and stores them
    into the cache, or removes tiles.

    :param pre_store_filter: a list with filter. each filter will be called
        with a tile before it will be stored to disc. the filter should
        return this or a new tile object.
    """

    def __init__(self, grid, cache, sources, format, locker, image_opts=None, request_format=None,
                 meta_buffer=None, meta_size=None, minimize_meta_requests=False, identifier=None,
                 pre_store_filter=None, concurrent_tile_creators=1, tile_creator_class=None,
                 bulk_meta_tiles=False,
                 rescale_tiles=0,
                 cache_rescaled_tiles=False,
                 dimensions=None
                 ):
        self.grid = grid
        self.cache = cache
        self.locker = locker
        self.identifier = identifier
        self.meta_grid = None
        self.format = format
        self.image_opts = image_opts
        self.request_format = request_format or format
        self.sources = sources
        self.minimize_meta_requests = minimize_meta_requests
        self._expire_timestamp = None
        self._refresh_before = {}
        self.pre_store_filter = pre_store_filter or []
        self.concurrent_tile_creators = concurrent_tile_creators
        self.tile_creator_class = tile_creator_class or TileCreator
        self.dimensions = dimensions

        self.rescale_tiles = rescale_tiles
        self.cache_rescaled_tiles = cache_rescaled_tiles

        if meta_buffer or (meta_size and not meta_size == [1, 1]):
            if all(source.supports_meta_tiles for source in sources):
                self.meta_grid = MetaGrid(grid, meta_size=meta_size, meta_buffer=meta_buffer)
            elif any(source.supports_meta_tiles for source in sources):
                raise ValueError('meta tiling configured but not supported by all sources')
            elif meta_size and not meta_size == [1, 1] and bulk_meta_tiles:
                # meta tiles configured but all sources are tiled
                # use bulk_meta_tile mode that download tiles in parallel
                self.meta_grid = MetaGrid(grid, meta_size=meta_size, meta_buffer=0)
                self.tile_creator_class = partial(self.tile_creator_class, bulk_meta_tiles=True)

    @contextmanager
    def session(self):
        """
        Context manager for access to the cache. Cleans up after usage
        for connection based caches.

        >>> with tile_manager.session(): #doctest: +SKIP
        ...    tile_manager.load_tile_coords(tile_coords)

        """
        yield
        self.cleanup()

    def cleanup(self):
        if hasattr(self.cache, 'cleanup'):
            self.cache.cleanup()

    def load_tile_coord(self, tile_coord, dimensions=None, with_metadata=False):
        return self.load_tile_coords(
            [tile_coord], dimensions=dimensions, with_metadata=with_metadata,
        )[0]

    def load_tile_coords(self, tile_coords, dimensions=None, with_metadata=False):
        tiles = TileCollection(tile_coords)
        rescale_till_zoom = 0
        if self.rescale_tiles:
            for t in tiles.tiles:
                # Use zoom level from first None tile.
                if t.coord is not None:
                    rescale_till_zoom = t.coord[2] + self.rescale_tiles
                    break
            else:
                return tiles

            if rescale_till_zoom < 0:
                rescale_till_zoom = 0
            if rescale_till_zoom > self.grid.levels:
                rescale_till_zoom = self.grid.levels

        # Remove tiles that are not in the cache coverage
        if self.cache.coverage:
            for t in tiles.tiles:
                if t.coord:
                    tile_bbox = self.grid.tile_bbox(t.coord)
                    if not self.cache.coverage.intersects(tile_bbox, self.grid.srs):
                        t.coord = None

        tiles = self._load_tile_coords(
            tiles, dimensions=dimensions, with_metadata=with_metadata,
            rescale_till_zoom=rescale_till_zoom, rescaled_tiles={},
        )

        for t in tiles.tiles:
            # Clip tiles if clipping is enabled for coverage
            if t.coord and self.cache.coverage and self.cache.coverage.clip and t.source:
                tile_bbox = self.grid.tile_bbox(t.coord)
                coverage = self.cache.coverage

                if coverage.intersects(tile_bbox, self.grid.srs):
                    t.source = mask_image_source_from_coverage(
                        t.source, tile_bbox, self.grid.srs, coverage, ImageOptions(transparent=True, format='png'))

            # Remove our internal marker source, for missing tiles.
            if t.source is RESCALE_TILE_MISSING:
                t.source = None
        return tiles

    def _is_tile_missing(self, tile, cache_only, dimensions=None):
        if tile.coord is None:
            return False
        if cache_only:
            # in cache_only mode, we already fetched the tile from cache
            return tile.is_missing()
        else:
            # missing or staled
            return not self.is_cached(tile, dimensions=dimensions)

    def _load_tile_coords(self, tiles, dimensions=None, with_metadata=False,
                          rescale_till_zoom=None, rescaled_tiles=None
                          ):
        uncached_tiles = []

        if rescaled_tiles:
            for t in tiles:
                if t.coord in rescaled_tiles:
                    t.source = rescaled_tiles[t.coord].source

        # load all in batch
        self.cache.load_tiles(tiles, with_metadata, dimensions=dimensions)

        # if no real source, we are running in cache_only mode
        cache_only = self.sources == [] or (len(self.sources) == 1 and isinstance(self.sources[0], DummySource))
        # if no rescale_tiles and cache_only, we dont have any additional processing to do
        if (self.rescale_tiles == 0 and cache_only):
            return tiles

        for tile in tiles:
            if self._is_tile_missing(tile, cache_only, dimensions=dimensions):
                uncached_tiles.append(tile)

        if uncached_tiles:
            creator = self.creator(dimensions=dimensions)
            created_tiles = creator.create_tiles(uncached_tiles)
            if not created_tiles and self.rescale_tiles:
                created_tiles = [self._scaled_tile(t, rescale_till_zoom, rescaled_tiles) for t in uncached_tiles]

            for created_tile in created_tiles:
                if created_tile.coord in tiles:
                    tiles[created_tile.coord].source = created_tile.source

        return tiles

    def remove_tile_coords(self, tile_coords, dimensions=None):
        tiles = TileCollection(tile_coords)
        self.cache.remove_tiles(tiles)

    def creator(self, dimensions=None):
        return self.tile_creator_class(self, dimensions=dimensions)

    def lock(self, tile):
        if self.meta_grid:
            tile = Tile(self.meta_grid.main_tile(tile.coord))
        return self.locker.lock(tile)

    def is_cached(self, tile, dimensions=None):
        """
        Return True if the tile is cached.
        """
        if isinstance(tile, tuple):
            tile = Tile(tile)
        if tile.coord is None:
            return True
        cached = self.cache.is_cached(tile, dimensions=dimensions)
        max_mtime = self.expire_timestamp(tile)
        if cached and max_mtime is not None:
            self.cache.load_tile_metadata(tile, dimensions=self.dimensions)
            # file time stamp must be rounded to integer since time conversion functions
            # mktime and timetuple strip decimals from seconds
            stale = int(tile.timestamp) <= max_mtime
            if stale:
                cached = False
        return cached

    def is_stale(self, tile, dimensions=None):
        """
        Return True if tile exists _and_ is expired.
        """
        if isinstance(tile, tuple):
            tile = Tile(tile)
        if self.cache.is_cached(tile, dimensions=dimensions):
            # tile exists
            if not self.is_cached(tile, dimensions=dimensions):
                # expired
                return True
            return False
        return False

    def expire_timestamp(self, tile=None):
        """
        Return the timestamp until which a tile should be accepted as up-to-date,
        or ``None`` if the tiles should not expire.

        :note: Returns _expire_timestamp by default.
        """
        if self._refresh_before:
            from mapproxy.seed.config import before_timestamp_from_options
            return before_timestamp_from_options(self._refresh_before)
        return self._expire_timestamp

    def apply_tile_filter(self, tile):
        """
        Apply all `pre_store_filter` to this tile.
        Returns filtered tile.
        """
        if tile.stored:
            return tile

        for img_filter in self.pre_store_filter:
            tile = img_filter(tile)
        return tile

    def _scaled_tile(self, tile, stop_zoom, rescaled_tiles):
        """
        Try to load tile by loading, scaling and clipping tiles from zoom levels above or
        below. stop_zoom determines if tiles from above should be scaled up, or if tiles
        from below should be scaled down.
        Returns an empty Tile if tile zoom level is stop_zoom.
        """
        if tile.coord in rescaled_tiles:
            return rescaled_tiles[tile.coord]

        # Cache tile in rescaled_tiles. We initially set source to a fixed
        # BlankImageSource and overwrite it if we actually rescaled the tile.
        tile.source = RESCALE_TILE_MISSING
        rescaled_tiles[tile.coord] = tile

        tile_bbox = self.grid.tile_bbox(tile.coord)
        current_zoom = tile.coord[2]
        if stop_zoom == current_zoom:
            return tile
        if stop_zoom > current_zoom:
            src_level = current_zoom + 1
        else:
            src_level = current_zoom - 1

        src_bbox, src_tile_grid, affected_tile_coords = self.grid.get_affected_level_tiles(tile_bbox, src_level)

        affected_tiles = TileCollection(affected_tile_coords)
        for t in affected_tiles:
            # Add sources of cached tiles, to avoid loading same tile multiple times
            # loading recursive.
            if t.coord in rescaled_tiles:
                t.source = rescaled_tiles[t.coord].source

        tile_collection = self._load_tile_coords(
            affected_tiles,
            rescale_till_zoom=stop_zoom,
            rescaled_tiles=rescaled_tiles,
        )

        if tile_collection.blank:
            return tile

        tile_sources = []
        for t in tile_collection:
            # Replace RESCALE_TILE_MISSING with None, before transforming tiles.
            tile_sources.append(t.source if t.source is not RESCALE_TILE_MISSING else None)

        tiled_image = TiledImage(tile_sources, src_bbox=src_bbox, src_srs=self.grid.srs,
                                 tile_grid=src_tile_grid, tile_size=self.grid.tile_size)
        tile.source = tiled_image.transform(tile_bbox, self.grid.srs, self.grid.tile_size, self.image_opts)

        if self.cache_rescaled_tiles:
            self.cache.store_tile(tile)
        return tile


# RESCALE_TILE_MISSING is a dummy source to prevent a tile cache from loading
# a tile that we already found out is missing.
RESCALE_TILE_MISSING = BlankImageSource((256, 256), ImageOptions())


class TileCreator(object):
    def __init__(self, tile_mgr, dimensions=None, image_merger=None, bulk_meta_tiles=False):
        self.cache = tile_mgr.cache
        self.sources = tile_mgr.sources
        self.grid = tile_mgr.grid
        self.meta_grid = tile_mgr.meta_grid
        self.bulk_meta_tiles = bulk_meta_tiles
        self.tile_mgr = tile_mgr
        self.dimensions = dimensions
        self.image_merger = image_merger

    def is_cached(self, tile, dimensions=None):
        """
        Return True if the tile is cached.
        """
        return self.tile_mgr.is_cached(tile, dimensions=dimensions)

    def is_stale(self, tile):
        """
        Return True if the tile exists in cache and is expired.
        """
        return self.tile_mgr.is_stale(tile)

    def create_tiles(self, tiles):
        if not self.sources:
            return []
        if not self.meta_grid:
            created_tiles = self._create_single_tiles(tiles)
        elif self.tile_mgr.minimize_meta_requests and len(tiles) > 1:
            # use minimal requests only for mulitple tile requests (ie not for TMS)
            meta_tile = self.meta_grid.minimal_meta_tile([t.coord for t in tiles])
            created_tiles = self._create_meta_tile(meta_tile)
        else:
            meta_tiles = []
            meta_bboxes = set()
            for tile in tiles:
                meta_tile = self.meta_grid.meta_tile(tile.coord)
                if meta_tile.bbox not in meta_bboxes:
                    meta_tiles.append(meta_tile)
                    meta_bboxes.add(meta_tile.bbox)

            created_tiles = self._create_meta_tiles(meta_tiles)

        return created_tiles

    def _create_single_tiles(self, tiles):
        if self.tile_mgr.concurrent_tile_creators > 1 and len(tiles) > 1:
            return self._create_threaded(self._create_single_tile, tiles)

        created_tiles = []
        for tile in tiles:
            created_tiles.extend(self._create_single_tile(tile))
        return created_tiles

    def _create_threaded(self, create_func, tiles):
        result = []
        async_pool = async_.Pool(self.tile_mgr.concurrent_tile_creators)
        for new_tiles in async_pool.imap(create_func, tiles):
            result.extend(new_tiles)
        return result

    def _create_single_tile(self, tile, dimensions=None):
        tile_bbox = self.grid.tile_bbox(tile.coord)
        query = MapQuery(tile_bbox, self.grid.tile_size, self.grid.srs,
                         self.tile_mgr.request_format, dimensions=self.dimensions)
        with self.tile_mgr.lock(tile):
            if not self.is_cached(tile, dimensions=dimensions):
                source = None
                try:
                    source = self._query_sources(query)
                # if source is not available, try to serve tile in cache
                except SourceError as e:
                    if self.is_stale(tile):
                        self.cache.load_tile(tile)
                    else:
                        reraise_exception(e, sys.exc_info())
                if not source:
                    return []
                if source.authorize_stale and self.is_stale(tile):
                    # The configuration authorises blank tiles generated by the error_handler
                    # to be replaced by stale tiles from cache.
                    self.cache.load_tile(tile)
                    return [tile]
                if self.tile_mgr.image_opts != source.image_opts:
                    # call as_buffer to force conversion into cache format
                    source.as_buffer(self.tile_mgr.image_opts)
                source.image_opts = self.tile_mgr.image_opts
                tile.source = source
                tile.cacheable = source.cacheable
                tile = self.tile_mgr.apply_tile_filter(tile)
                if source.cacheable:
                    self.cache.store_tile(tile)
            else:
                self.cache.load_tile(tile)
        return [tile]

    def _query_sources(self, query):
        """
        Query all sources and return the results as a single ImageSource.
        Multiple sources will be merged into a single image.
        """

        # directly return get_map without merge if ...
        if (len(self.sources) == 1 and
                not self.image_merger and  # no special image_merger (like BandMerger)
                not (self.sources[0].coverage and  # no clipping coverage
                     self.sources[0].coverage.clip and
                     self.sources[0].coverage.intersects(query.bbox, query.srs))):
            try:
                return self.sources[0].get_map(query)
            except BlankImage:
                return None

        def get_map_from_source(source):
            try:
                img = source.get_map(query)
            except BlankImage:
                return None, None
            else:
                return (img, source.coverage)

        layers = []
        for layer in async_.imap(get_map_from_source, self.sources):
            if layer[0] is not None:
                layers.append(layer)

        return merge_images(layers, size=query.size, bbox=query.bbox, bbox_srs=query.srs,
                            image_opts=self.tile_mgr.image_opts, merger=self.image_merger)

    def _create_meta_tiles(self, meta_tiles):
        if self.bulk_meta_tiles:
            created_tiles = []
            for meta_tile in meta_tiles:
                created_tiles.extend(self._create_bulk_meta_tile(meta_tile))
            return created_tiles

        if self.tile_mgr.concurrent_tile_creators > 1 and len(meta_tiles) > 1:
            return self._create_threaded(self._create_meta_tile, meta_tiles)

        created_tiles = []
        for meta_tile in meta_tiles:
            created_tiles.extend(self._create_meta_tile(meta_tile))
        return created_tiles

    def _create_meta_tile(self, meta_tile):
        """
        _create_meta_tile queries a single meta tile and splits it into
        tiles.
        """
        tile_size = self.grid.tile_size
        query = MapQuery(meta_tile.bbox, meta_tile.size, self.grid.srs, self.tile_mgr.request_format,
                         dimensions=self.dimensions)
        main_tile = Tile(meta_tile.main_tile_coord)
        with self.tile_mgr.lock(main_tile):
            if not all(self.is_cached(t, dimensions=self.dimensions) for t in meta_tile.tiles if t is not None):
                meta_tile_image = self._query_sources(query)
                if not meta_tile_image:
                    return []
                splitted_tiles = split_meta_tiles(meta_tile_image, meta_tile.tile_patterns,
                                                  tile_size, self.tile_mgr.image_opts)
                splitted_tiles = [self.tile_mgr.apply_tile_filter(t) for t in splitted_tiles]
                if meta_tile_image.cacheable:
                    self.cache.store_tiles(splitted_tiles, dimensions=self.dimensions)
                return splitted_tiles
            # else
        tiles = [Tile(coord) for coord in meta_tile.tiles]
        self.cache.load_tiles(tiles, dimensions=self.dimensions)
        return tiles

    def _create_bulk_meta_tile(self, meta_tile):
        """
        _create_bulk_meta_tile queries each tile of the meta tile in parallel
        (using concurrent_tile_creators).
        """
        tile_size = self.grid.tile_size
        main_tile = Tile(meta_tile.main_tile_coord)
        with self.tile_mgr.lock(main_tile):
            if not all(self.is_cached(t, dimensions=self.dimensions) for t in meta_tile.tiles if t is not None):
                async_pool = async_.Pool(self.tile_mgr.concurrent_tile_creators)

                def query_tile(coord):
                    try:
                        query = MapQuery(
                            self.grid.tile_bbox(coord), tile_size, self.grid.srs, self.tile_mgr.request_format,
                            dimensions=self.dimensions)
                        tile_image = self._query_sources(query)
                        if tile_image is None:
                            return None

                        if self.tile_mgr.image_opts != tile_image.image_opts:
                            # call as_buffer to force conversion into cache format
                            tile_image.as_buffer(self.tile_mgr.image_opts)

                        tile = Tile(coord, cacheable=tile_image.cacheable)
                        tile.source = tile_image
                        tile = self.tile_mgr.apply_tile_filter(tile)
                    except BlankImage:
                        return None
                    else:
                        return tile

                tiles = []
                for tile_task in async_pool.imap(query_tile,
                                                 [t for t in meta_tile.tiles if t is not None],
                                                 use_result_objects=True,
                                                 ):
                    if tile_task.exception is None:
                        tile = tile_task.result
                        if tile is not None:
                            tiles.append(tile)
                    else:
                        ex = tile_task.exception
                        async_pool.shutdown(True)
                        reraise(ex)

                self.cache.store_tiles([t for t in tiles if t.cacheable], dimensions=self.dimensions)
                return tiles

            # else
        tiles = [Tile(coord) for coord in meta_tile.tiles]
        self.cache.load_tiles(tiles, dimensions=self.dimensions)
        return tiles


class Tile(object):
    """
    Internal data object for all tiles. Stores the tile-``coord`` and the tile data.

    :ivar source: the data of this tile
    :type source: ImageSource
    """

    def __init__(self, coord, source=None, cacheable=True):
        self.coord = coord
        self.source = source
        self.location = None
        self.stored = False
        self._cacheable = cacheable
        self.size = None
        self.timestamp = None

    def _cacheable_get(self):
        return CacheInfo(cacheable=self._cacheable, timestamp=self.timestamp,
                         size=self.size)

    def _cacheable_set(self, cacheable):
        if isinstance(cacheable, bool):
            self._cacheable = cacheable
        else:  # assume cacheable is CacheInfo
            self._cacheable = cacheable.cacheable
            self.timestamp = cacheable.timestamp
            self.size = cacheable.size

    cacheable = property(_cacheable_get, _cacheable_set)

    def source_buffer(self, *args, **kw):
        if self.source is not None:
            return self.source.as_buffer(*args, **kw)
        else:
            return None

    def source_image(self, *args, **kw):
        if self.source is not None:
            return self.source.as_image(*args, **kw)
        else:
            return None

    def is_missing(self):
        """
        Returns ``True`` when the tile has no ``data``, except when the ``coord``
        is ``None``. It doesn't check if the tile exists.

        >>> Tile((1, 2, 3)).is_missing()
        True
        >>> Tile((1, 2, 3), './tmp/foo').is_missing()
        False
        >>> Tile(None).is_missing()
        False
        """
        if self.coord is None:
            return False
        return self.source is None

    def __eq__(self, other):
        """
        >>> Tile((0, 0, 1)) == Tile((0, 0, 1))
        True
        >>> Tile((0, 0, 1)) == Tile((1, 0, 1))
        False
        >>> Tile((0, 0, 1)) is None
        False
        """
        if isinstance(other, Tile):
            return (self.coord == other.coord and
                    self.source == other.source)
        else:
            return NotImplemented

    def __ne__(self, other):
        """
        >>> Tile((0, 0, 1)) != Tile((0, 0, 1))
        False
        >>> Tile((0, 0, 1)) != Tile((1, 0, 1))
        True
        >>> Tile((0, 0, 1)) != None
        True
        """
        equal_result = self.__eq__(other)
        if equal_result is NotImplemented:
            return NotImplemented
        else:
            return not equal_result

    def __repr__(self):
        return 'Tile(%r, source=%r)' % (self.coord, self.source)


class CacheInfo(object):
    def __init__(self, cacheable=True, timestamp=None, size=None):
        self.cacheable = cacheable
        self.timestamp = timestamp
        self.size = size

    def __bool__(self):
        return self.cacheable


class TileCollection(object):
    def __init__(self, tile_coords):
        self.tiles = [Tile(coord) for coord in tile_coords]
        self.tiles_dict = {}
        for tile in self.tiles:
            self.tiles_dict[tile.coord] = tile

    def __getitem__(self, idx_or_coord):
        if isinstance(idx_or_coord, int):
            return self.tiles[idx_or_coord]
        if idx_or_coord in self.tiles_dict:
            return self.tiles_dict[idx_or_coord]
        return Tile(idx_or_coord)

    def __contains__(self, tile_or_coord):
        if isinstance(tile_or_coord, tuple):
            return tile_or_coord in self.tiles_dict
        if hasattr(tile_or_coord, 'coord'):
            return tile_or_coord.coord in self.tiles_dict
        return False

    def __len__(self):
        return len(self.tiles)

    def __iter__(self):
        return iter(self.tiles)

    @property
    def empty(self):
        """
        Returns True if no tile in this collection contains a source.
        """
        return all((t.source is None for t in self.tiles))

    @property
    def blank(self):
        """
        Returns True if all sources collection are BlankImageSources or have not source at all.
        """
        return all((t.source is None or isinstance(t.source, BlankImageSource) for t in self.tiles))

    def __repr__(self):
        return 'TileCollection(%r)' % self.tiles


def split_meta_tiles(meta_tile, tiles, tile_size, image_opts):
    try:
        # TODO png8
        # if not self.transparent and format == 'png':
        #     format = 'png8'
        splitter = TileSplitter(meta_tile, image_opts)
    except IOError:
        # TODO
        raise
    split_tiles = []
    for tile in tiles:
        tile_coord, crop_coord = tile
        if tile_coord is None:
            continue
        data = splitter.get_tile(crop_coord, tile_size)
        new_tile = Tile(tile_coord, cacheable=meta_tile.cacheable)
        new_tile.source = data
        split_tiles.append(new_tile)
    return split_tiles



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import io
import os
import shutil

import pytest

from mapproxy.multiapp import make_wsgi_app

from mapproxy.test.system import SysTest, WSGITestApp


class TestMultiapp(SysTest):

    @pytest.fixture(scope="class")
    def app(self, base_dir):
        app = make_wsgi_app(base_dir.strpath, allow_listing=False)
        return WSGITestApp(app, use_unicode=False)

    @pytest.fixture(scope="class")
    def base_dir(self, tmpdir_factory):
        dir = tmpdir_factory.mktemp("base_dir")

        fixture_dir = os.path.join(os.path.dirname(__file__), "fixture")
        shutil.copy(os.path.join(fixture_dir, "multiapp1.yaml"), dir.strpath)
        shutil.copy(os.path.join(fixture_dir, "multiapp2.yaml"), dir.strpath)

        return dir

    def test_index_without_list(self, app):
        resp = app.get("/")
        assert "MapProxy" in resp
        assert "multiapp1" not in resp

    def test_index_with_list(self, app):
        try:
            app.app.list_apps = True
            resp = app.get("/")
            assert "MapProxy" in resp
            assert "multiapp1" in resp
        finally:
            app.app.list_apps = False

    def test_unknown_app(self, app):
        app.get("/unknownapp", status=404)
        # assert status == 404 Not Found in app.get

    def test_known_app(self, app):
        resp = app.get("/multiapp1")
        assert "demo" in resp

    def test_reloading(self, app, base_dir):
        resp = app.get("/multiapp1")
        assert "demo" in resp
        app_config = base_dir.join("multiapp1.yaml").strpath

        replace_text_in_file(app_config, "  demo:", "  #demo:", ts_delta=5)

        resp = app.get("/multiapp1")
        assert "demo" not in resp

        replace_text_in_file(app_config, "  #demo:", "  demo:", ts_delta=10)

        resp = app.get("/multiapp1")
        assert "demo" in resp


def replace_text_in_file(filename, old, new, ts_delta=2):
    text = io.open(filename, encoding="utf-8").read()
    text = text.replace(old, new)
    io.open(filename, "w", encoding="utf-8").write(text)

    # file timestamps are not precise enough (1sec)
    # add larger delta to force reload
    m_time = os.path.getmtime(filename)
    os.utime(filename, (m_time + ts_delta, m_time + ts_delta))



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

from io import BytesIO

import pytest

from mapproxy.request.wms import WMS111MapRequest, WMS111FeatureInfoRequest
from mapproxy.test.image import tmp_image, check_format
from mapproxy.test.http import mock_httpd
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "formats.yaml"


def assert_file_format(filename, format):
    assert filename.check()
    check_format(filename.read_binary(), format)


class TestWMS111(SysTest):

    def setup_method(self):
        self.common_req = WMS111MapRequest(
            url="/service?", param=dict(service="WMS", version="1.1.1")
        )
        self.common_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="0,0,180,80",
                width="200",
                height="200",
                layers="wms_cache",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
            ),
        )
        self.common_direct_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="0,0,10,10",
                width="200",
                height="200",
                layers="wms_cache",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
            ),
        )
        self.common_fi_req = WMS111FeatureInfoRequest(
            url="/service?",
            param=dict(
                x="10",
                y="20",
                width="200",
                height="200",
                layers="wms_cache",
                format="image/png",
                query_layers="wms_cache",
                styles="",
                bbox="1000,400,2000,1400",
                srs="EPSG:900913",
            ),
        )
        self.expected_base_path = "/service?SERVICE=WMS&REQUEST=GetMap&HEIGHT=256" "&SRS=EPSG%3A900913&styles=&VERSION=1.1.1&WIDTH=256" "&BBOX=0.0,0.0,20037508.3428,20037508.3428"  # noqa
        self.expected_direct_base_path = "/service?SERVICE=WMS&REQUEST=GetMap&HEIGHT=200" "&SRS=EPSG%3A4326&styles=&VERSION=1.1.1&WIDTH=200" "&BBOX=0.0,0.0,10.0,10.0"  # noqa

    @pytest.mark.parametrize(
        "layer,source,wms_format,cache_format,req_format",
        [
            ["jpeg_cache_tiff_source", "tiffsource", "png", "jpeg", "tiff"],
            ["jpeg_cache_tiff_source", "tiffsource", "jpeg", "jpeg", "tiff"],
            ["jpeg_cache_tiff_source", "tiffsource", "tiff", "jpeg", "tiff"],
            ["jpeg_cache_tiff_source", "tiffsource", "gif", "jpeg", "tiff"],
            ["png_cache_all_source", "allsource", "png", "png", "png"],
            ["png_cache_all_source", "allsource", "jpeg", "png", "png"],
            ["jpeg_cache_png_jpeg_source", "pngjpegsource", "jpeg", "jpeg", "jpeg"],
            ["jpeg_cache_png_jpeg_source", "pngjpegsource", "png", "jpeg", "jpeg"],
        ],
    )
    def test_get_cached(
        self, app, cache_dir, layer, source, wms_format, cache_format, req_format
    ):
        with tmp_image((256, 256), format=req_format) as img:
            expected_req = (
                {
                    "path": self.expected_base_path
                    + "&layers="
                    + source
                    + "&format=image%2F"
                    + req_format
                },
                {
                    "body": img.read(),
                    "headers": {"content-type": "image/" + req_format},
                },
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                self.common_map_req.params["layers"] = layer
                self.common_map_req.params["format"] = "image/" + wms_format
                resp = app.get(self.common_map_req)
                assert resp.content_type == "image/" + wms_format
                check_format(BytesIO(resp.body), wms_format)
        assert_file_format(
            cache_dir.join(
                layer + "_EPSG900913/01/000/000/001/000/000/001." + cache_format
            ),
            cache_format,
        )

    @pytest.mark.parametrize(
        "layer,source,wms_format,req_format",
        [
            ["jpeg_cache_tiff_source", "tiffsource", "gif", "tiff"],
            ["jpeg_cache_tiff_source", "tiffsource", "jpeg", "tiff"],
            ["jpeg_cache_tiff_source", "tiffsource", "png", "tiff"],
            ["png_cache_all_source", "allsource", "gif", "gif"],
            ["png_cache_all_source", "allsource", "png", "png"],
            ["png_cache_all_source", "allsource", "tiff", "tiff"],
            ["jpeg_cache_png_jpeg_source", "pngjpegsource", "jpeg", "jpeg"],
            ["jpeg_cache_png_jpeg_source", "pngjpegsource", "png", "png"],
            ["jpeg_cache_png_jpeg_source", "pngjpegsource", "tiff", "png"],
            ["jpeg_cache_png_jpeg_source", "pngjpegsource", "gif", "png"],
        ],
    )
    def test_get_direct(self, app, layer, source, wms_format, req_format):
        with tmp_image((256, 256), format=req_format) as img:
            expected_req = (
                {
                    "path": self.expected_direct_base_path
                    + "&layers="
                    + source
                    + "&format=image%2F"
                    + req_format
                },
                {
                    "body": img.read(),
                    "headers": {"content-type": "image/" + req_format},
                },
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                self.common_direct_map_req.params["layers"] = layer
                self.common_direct_map_req.params["format"] = "image/" + wms_format
                resp = app.get(self.common_direct_map_req)
                assert resp.content_type == "image/" + wms_format
                check_format(BytesIO(resp.body), wms_format)


class TestTMS(SysTest):

    def setup_method(self):
        self.expected_base_path = "/service?SERVICE=WMS&REQUEST=GetMap&HEIGHT=256" "&SRS=EPSG%3A900913&styles=&VERSION=1.1.1&WIDTH=256" "&BBOX=0.0,0.0,20037508.3428,20037508.3428"  # noqa
        self.expected_direct_base_path = "/service?SERVICE=WMS&REQUEST=GetMap&HEIGHT=200" "&SRS=EPSG%3A4326&styles=&VERSION=1.1.1&WIDTH=200" "&BBOX=0.0,0.0,10.0,10.0"  # noqa

    @pytest.mark.parametrize(
        "layer,source,tms_format,cache_format,req_format",
        [
            ["jpeg_cache_tiff_source", "tiffsource", "jpeg", "jpeg", "tiff"],
            ["png_cache_all_source", "allsource", "png", "png", "png"],
            ["jpeg_cache_png_jpeg_source", "pngjpegsource", "jpeg", "jpeg", "jpeg"],
        ],
    )
    def test_get_cached(
        self, app, cache_dir, layer, source, tms_format, cache_format, req_format
    ):
        with tmp_image((256, 256), format=req_format) as img:
            expected_req = (
                {
                    "path": self.expected_base_path
                    + "&layers="
                    + source
                    + "&format=image%2F"
                    + req_format
                },
                {
                    "body": img.read(),
                    "headers": {"content-type": "image/" + req_format},
                },
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tms/1.0.0/%s/0/1/1.%s" % (layer, tms_format))
                assert resp.content_type == "image/" + tms_format
                # check_format(BytesIO(resp.body), tms_format)
        assert_file_format(
            cache_dir.join(
                layer + "_EPSG900913/01/000/000/001/000/000/001." + cache_format
            ),
            cache_format,
        )



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
WMS service handler
"""
from __future__ import print_function

import re
from functools import partial
import logging
from collections import OrderedDict

from mapproxy.request.wmts import (
    wmts_request, make_wmts_rest_request_parser,
    URLTemplateConverter,
    FeatureInfoURLTemplateConverter,
)
from mapproxy.layer import InfoQuery
from mapproxy.featureinfo import combine_docs
from mapproxy.service.base import Server
from mapproxy.response import Response
from mapproxy.exception import RequestError
from mapproxy.util.coverage import load_limited_to

from mapproxy.template import template_loader, bunch

env = {'bunch': bunch}
get_template = template_loader(__package__, 'templates', namespace=env)
log = logging.getLogger(__name__)


class WMTSServer(Server):
    service = 'wmts'

    def __init__(self, layers, md, request_parser=None, max_tile_age=None, info_formats=None):
        Server.__init__(self)
        self.request_parser = request_parser or wmts_request
        self.md = md
        self.max_tile_age = max_tile_age
        self.layers, self.matrix_sets = self._matrix_sets(layers)
        self.capabilities_class = Capabilities
        self.fi_transformers = None
        self.info_formats = info_formats or {}

    def _matrix_sets(self, layers):
        sets = {}
        layers_grids = OrderedDict()
        for layer in layers.values():
            grid = layer.grid
            if not grid.supports_access_with_origin('nw'):
                log.warning("skipping layer '%s' for WMTS, grid '%s' of cache '%s' is not compatible with WMTS",
                            layer.name, grid.name, layer.md['cache_name'])
                continue
            if grid.name not in sets:
                try:
                    sets[grid.name] = TileMatrixSet(grid)
                except AssertionError:
                    continue  # TODO
            layers_grids.setdefault(layer.name, OrderedDict())[grid.name] = layer
        wmts_layers = OrderedDict()
        for layer_name, layers in layers_grids.items():
            wmts_layers[layer_name] = WMTSTileLayer(layers)
        return wmts_layers, sets.values()

    def capabilities(self, request):
        service = self._service_md(request)
        layers = self.authorized_tile_layers(request.http.environ)

        result = self.capabilities_class(service, layers, self.matrix_sets,
                                         info_formats=self.info_formats).render(request)
        return Response(result, mimetype='application/xml')

    def tile(self, request):
        self.check_request(request)

        tile_layer = self.layers[request.layer][request.tilematrixset]
        if not request.format:
            request.format = tile_layer.format

        self.check_request_dimensions(tile_layer, request)

        limited_to = self.authorize_tile_layer(tile_layer, request)

        def decorate_img(image):
            query_extent = tile_layer.grid.srs.srs_code, tile_layer.tile_bbox(request)
            return self.decorate_img(image, 'wmts', [tile_layer.name], request.http.environ, query_extent)

        tile = tile_layer.render(request, coverage=limited_to, decorate_img=decorate_img)

        # set the content_type to tile.format and not to request.format ( to support mixed_mode)
        resp = Response(tile.as_buffer(), content_type='image/' + tile.format)
        resp.cache_headers(tile.timestamp, etag_data=(tile.timestamp, tile.size),
                           max_age=self.max_tile_age)
        resp.make_conditional(request.http)
        return resp

    def featureinfo(self, request):
        infos = []
        self.check_request(request, self.info_formats)

        tile_layer = self.layers[request.layer][request.tilematrixset]
        if not request.format:
            request.format = tile_layer.format

        feature_count = None
        # WMTS REST style request do not have request params
        if hasattr(request, 'params'):
            feature_count = request.params.get('feature_count', None)

        bbox = tile_layer.grid.tile_bbox(request.tile)
        query = InfoQuery(bbox, tile_layer.grid.tile_size, tile_layer.grid.srs, request.pos,
                          request.infoformat, feature_count=feature_count)
        self.check_request_dimensions(tile_layer, request)
        coverage = self.authorize_tile_layer(tile_layer, request, featureinfo=True)

        if not tile_layer.info_sources:
            raise RequestError('layer %s not queryable' % str(request.layer),
                               code='OperationNotSupported', request=request)

        if coverage and not coverage.contains(query.coord, query.srs):
            infos = []
        else:
            for source in tile_layer.info_sources:
                info = source.get_info(query)
                if info is None:
                    continue
                infos.append(info)

        mimetype = request.infoformat

        if not infos:
            return Response('', mimetype=mimetype)

        resp, _ = combine_docs(infos)

        return Response(resp, mimetype=mimetype)

    def authorize_tile_layer(self, tile_layer, request, featureinfo=False):
        if 'mapproxy.authorize' not in request.http.environ:
            return

        query_extent = tile_layer.grid.srs.srs_code, tile_layer.tile_bbox(request)

        service = 'wmts'
        key = 'tile'
        if featureinfo:
            service += '.featureinfo'
            key = 'featureinfo'

        result = request.http.environ['mapproxy.authorize'](service, [tile_layer.name],
                                                            query_extent=query_extent, environ=request.http.environ)
        if result['authorized'] == 'unauthenticated':
            raise RequestError('unauthorized', status=401)
        if result['authorized'] == 'full':
            return
        if result['authorized'] == 'partial':
            if result['layers'].get(tile_layer.name, {}).get(key, False) is True:
                limited_to = result['layers'][tile_layer.name].get('limited_to')
                if not limited_to:
                    limited_to = result.get('limited_to')
                if limited_to:
                    return load_limited_to(limited_to)
                else:
                    return None
        raise RequestError('forbidden', status=403)

    def authorized_tile_layers(self, env):
        if 'mapproxy.authorize' in env:
            result = env['mapproxy.authorize']('wmts', [x for x in self.layers],
                                               query_extent=None, environ=env)
            if result['authorized'] == 'unauthenticated':
                raise RequestError('unauthorized', status=401)
            if result['authorized'] == 'full':
                return list(self.layers.values())
            if result['authorized'] == 'none':
                raise RequestError('forbidden', status=403)
            allowed_layers = []
            for layer in self.layers.values():
                if result['layers'].get(layer.name, {}).get('tile', False):
                    allowed_layers.append(layer)
            return allowed_layers
        else:
            return list(self.layers.values())

    def check_request(self, request, info_formats=None):
        request.make_request()
        if request.layer not in self.layers:
            raise RequestError('unknown layer: ' + str(request.layer),
                               code='InvalidParameterValue', request=request)
        if request.tilematrixset not in self.layers[request.layer]:
            raise RequestError('unknown tilematrixset: ' + str(request.tilematrixset),
                               code='InvalidParameterValue', request=request)

        if info_formats is not None:
            if '/' in request.infoformat:  # mimetype
                if request.infoformat not in self.info_formats.values():
                    raise RequestError('unknown infoformat: ' + str(request.infoformat),
                                       code='InvalidParameterValue', request=request)
            else:  # RESTful suffix
                if request.infoformat not in self.info_formats:
                    raise RequestError('unknown infoformat: ' + str(request.infoformat),
                                       code='InvalidParameterValue', request=request)
                # set mimetype as infoformat
                request.infoformat = self.info_formats[request.infoformat]

    def check_request_dimensions(self, tile_layer, request):
        # allow arbitrary dimensions in KVP service
        # actual used values are checked later in TileLayer
        pass

    def _service_md(self, tile_request):
        md = dict(self.md)
        md['url'] = tile_request.url
        return md


class WMTSRestServer(WMTSServer):
    """
    OGC WMTS 1.0.0 RESTful Server
    """
    service = None
    names = ('wmts',)
    request_methods = ('tile', 'capabilities')
    default_template = '/{Layer}/{TileMatrixSet}/{TileMatrix}/{TileCol}/{TileRow}.{Format}'
    default_info_template = '/{Layer}/{TileMatrixSet}/{TileMatrix}/{TileCol}/{TileRow}/{I}/{J}.{InfoFormat}'

    def __init__(self, layers, md, max_tile_age=None, template=None, fi_template=None, info_formats=None):
        WMTSServer.__init__(self, layers, md)
        self.max_tile_age = max_tile_age
        self.template = template or self.default_template
        self.fi_template = fi_template or self.default_info_template
        self.info_formats = info_formats or {}
        self.url_converter = URLTemplateConverter(self.template)
        self.fi_url_converter = FeatureInfoURLTemplateConverter(self.fi_template)
        self.request_parser = make_wmts_rest_request_parser(self.url_converter, self.fi_url_converter)
        self.capabilities_class = partial(
            RestfulCapabilities, url_converter=self.url_converter, fi_url_converter=self.fi_url_converter)

    def check_request_dimensions(self, tile_layer, request):
        # check that unknown dimension for this layer are set to default
        if request.dimensions:
            for dimension, value in request.dimensions.items():
                dimension = dimension.lower()
                if dimension not in tile_layer.dimensions and value != 'default':
                    raise RequestError('unknown dimension: ' + str(dimension),
                                       code='InvalidParameterValue', request=request)


class Capabilities(object):
    """
    Renders WMS capabilities documents.
    """

    def __init__(self, server_md, layers, matrix_sets, info_formats=None):
        self.service = server_md
        self.layers = layers
        self.info_formats = info_formats or {}
        self.matrix_sets = matrix_sets

    def render(self, _map_request):
        return self._render_template(_map_request.capabilities_template)

    def template_context(self):
        service = bunch(default='', **self.service)
        base_url = re.sub(r'/service$', '', service.url)
        legendurls = {}
        for layer in self.layers:
            if layer.md['wmts_kvp_legendurl'] is not None:
                legendurls[layer.name] = (
                    layer.md['wmts_kvp_legendurl']
                    .replace('{base_url}', base_url)
                    .replace('{layer_name}', layer.name)
                )
            else:
                legendurls[layer.name] = None

        return dict(service=service,
                    restful=False,
                    layers=self.layers,
                    info_formats=self.info_formats,
                    tile_matrix_sets=self.matrix_sets,
                    legendurls=legendurls)

    def _render_template(self, template):
        template = get_template(template)
        doc = template.substitute(**self.template_context())
        # strip blank lines
        doc = '\n'.join(x for x in doc.split('\n') if x.rstrip())
        return doc


class RestfulCapabilities(Capabilities):
    def __init__(self, server_md, layers, matrix_sets, url_converter, fi_url_converter, info_formats=None):
        Capabilities.__init__(self, server_md, layers, matrix_sets, info_formats=info_formats)
        self.url_converter = url_converter
        self.fi_url_converter = fi_url_converter

    def template_context(self):
        service = bunch(default='', **self.service)
        base_url = re.sub(r'/wmts$', '', service.url)
        legendurls = {}
        for layer in self.layers:
            if layer.md['wmts_rest_legendurl'] is not None:
                legendurls[layer.name] = (
                    layer.md['wmts_rest_legendurl']
                    .replace('{base_url}', base_url)
                    .replace('{layer_name}', layer.name)
                )
            else:
                legendurls[layer.name] = None

        return dict(service=service,
                    restful=True,
                    layers=self.layers,
                    info_formats=self.info_formats,
                    tile_matrix_sets=self.matrix_sets,
                    resource_template=self.url_converter.template,
                    fi_resource_template=self.fi_url_converter.template,
                    # dimension_key maps lowercase dimensions to the actual
                    # casing from the restful template
                    dimension_keys=dict((k.lower(), k) for k in self.url_converter.dimensions),
                    format_resource_template=format_resource_template,
                    format_info_resource_template=format_info_resource_template,
                    legendurls=legendurls
                    )


def format_resource_template(layer, template, service):
    if '{Format}' in template:
        template = template.replace('{Format}', layer.format)

    if '{Layer}' in template:
        template = template.replace('{Layer}', layer.name)

    return service.url + template


def format_info_resource_template(layer, template, info_format, service):
    if '{InfoFormat}' in template:
        template = template.replace('{InfoFormat}', info_format)

    if '{Layer}' in template:
        template = template.replace('{Layer}', layer.name)

    return service.url + template


class WMTSTileLayer(object):
    """
    Wrap multiple TileLayers for the same cache but with different grids.
    """

    def __init__(self, layers):
        self.grids = [lyr.grid for lyr in layers.values()]
        self.layers = layers
        self._layer = layers[list(layers.keys())[0]]

    def __getattr__(self, name):
        return getattr(self._layer, name)

    def __contains__(self, gridname):
        return gridname in self.layers

    def __getitem__(self, gridname):
        return self.layers[gridname]


# calculated from well-known scale set GoogleCRS84Quad
METERS_PER_DEEGREE = 111319.4907932736


def meter_per_unit(srs):
    if srs.is_latlong:
        return METERS_PER_DEEGREE
    return 1


class TileMatrixSet(object):
    def __init__(self, grid):
        self.grid = grid
        self.name = grid.name
        self.srs_name = grid.srs.srs_code
        self.tile_matrices = list(self._tile_matrices())

    def __iter__(self):
        return iter(self.tile_matrices)

    def _tile_matrices(self):
        for level, res in self.grid.resolutions.iteritems():
            origin = self.grid.origin_tile(level, 'ul')
            bbox = self.grid.tile_bbox(origin)
            topleft = bbox[0], bbox[3]
            if self.grid.srs.is_axis_order_ne:
                topleft = bbox[3], bbox[0]
            grid_size = self.grid.grid_sizes[level]
            scale_denom = round(res / (0.28 / 1000) * meter_per_unit(self.grid.srs), 10)
            yield bunch(
                identifier=level,
                topleft=topleft,
                grid_size=grid_size,
                scale_denom=f'{scale_denom}'.strip('0').strip('.'),
                tile_size=self.grid.tile_size,
            )



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
ctypes utilities.
"""
from __future__ import print_function

import sys
import os

from ctypes import CDLL
from ctypes.util import find_library as _find_library


default_locations = dict(
    darwin=dict(
        paths=['/opt/local/lib'],
        exts=['.dylib'],
    ),
    win32=dict(
        paths=[os.path.dirname(os.__file__) + '/../../../DLLs'],
        exts=['.dll']
    ),
    other=dict(
        paths=[],  # MAPPROXY_LIB_PATH will add paths here
        exts=['.so']
    ),
)

additional_lib_path = os.environ.get('MAPPROXY_LIB_PATH')
if additional_lib_path:
    additional_lib_path = additional_lib_path.split(os.pathsep)
    additional_lib_path.reverse()
    for locs in default_locations.values():
        for path in additional_lib_path:
            locs['paths'].insert(0, path)


def load_library(lib_names, locations_conf=default_locations):
    """
    Load the `lib_name` library with ctypes.
    If ctypes.util.find_library does not find the library,
    different path and filename extensions will be tried.

    Retruns the loaded library or None.
    """
    if isinstance(lib_names, str):
        lib_names = [lib_names]

    for lib_name in lib_names:
        lib = load_library_(lib_name, locations_conf)
        if lib is not None:
            return lib


def load_library_(lib_name, locations_conf=default_locations):
    lib_path = find_library(lib_name)

    if lib_path:
        return CDLL(lib_path)

    if sys.platform in locations_conf:
        paths = locations_conf[sys.platform]['paths']
        exts = locations_conf[sys.platform]['exts']
        lib_path = find_library(lib_name, paths, exts)
    else:
        paths = locations_conf['other']['paths']
        exts = locations_conf['other']['exts']
        lib_path = find_library(lib_name, paths, exts)

    if lib_path:
        return CDLL(lib_path)


def find_library(lib_name, paths=None, exts=None):
    """
    Search for library in all permutations of `paths` and `exts`.
    If nothing is found None is returned.
    """
    if not paths or not exts:
        lib = None
        try:
            lib = _find_library(lib_name)
        except FileNotFoundError:
            pass
        if lib is None and lib_name.startswith('lib'):
            try:
                lib = _find_library(lib_name[3:])
            except FileNotFoundError:
                pass
        return lib

    for lib_name in [lib_name] + ([lib_name[3:]] if lib_name.startswith('lib') else []):
        for path in paths:
            for ext in exts:
                lib_path = os.path.join(path, lib_name + ext)
                if os.path.exists(lib_path):
                    return lib_path

    return None


if __name__ == '__main__':
    print(load_library(sys.argv[1]))



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

from PIL import Image
from mapproxy.grid.tile_grid import tile_grid
from mapproxy.image import BlankImageSource
from mapproxy.image import ImageSource
from mapproxy.image.opts import ImageOptions
from mapproxy.layer import MapLayer, DefaultMapExtent
from mapproxy.service.base import Server
from mapproxy.service.tile import TileServer
from mapproxy.service.wms import WMSGroupLayer, WMSServer
from mapproxy.service.wmts import WMTSServer
from mapproxy.test.http import make_wsgi_env


class DummyLayer(MapLayer):
    transparent = True
    extent = DefaultMapExtent()
    has_legend = False
    queryable = False

    def __init__(self, name):
        MapLayer.__init__(self)
        self.name = name
        self.requested = False
        self.queried = False
        self.nominal_scale = None

    def get_map(self, query):
        self.requested = True

    def get_info(self, query):
        self.queried = True

    def map_layers_for_query(self, query):
        return [(self.name, self)]

    def info_layers_for_query(self, query):
        return [(self.name, self)]


class DummyTileLayer(object):
    def __init__(self, name):
        self.requested = False
        self.name = name
        self.grid = tile_grid(900913)

    def tile_bbox(self, request, use_profiles=False):
        # this dummy code does not handle profiles and different tile origins!
        return self.grid.tile_bbox(request.tile)

    def render(self, tile_request, use_profiles=None, coverage=None, decorate_img=None):
        self.requested = True
        resp = BlankImageSource((256, 256), image_opts=ImageOptions(format='image/png'))
        resp.timestamp = 0
        return resp


class TestDecorateImg(object):

    def setup_method(self):
        # Base server
        self.server = Server()
        # WMS Server
        root_layer = WMSGroupLayer(None, 'root layer', None, [DummyLayer('wms_cache')])
        self.wms_server = WMSServer(
            md={}, root_layer=root_layer, srs=['EPSG:4326'],
            image_formats={'image/png': ImageOptions(format='image/png')}
        )
        # Tile Servers
        layers = OrderedDict()
        layers["wms_cache_EPSG900913"] = DummyTileLayer('wms_cache')
        self.tile_server = TileServer(layers, {})
        self.wmts_server = WMTSServer(layers, {})
        # Common arguments
        self.query_extent = ('EPSG:27700', (0, 0, 700000, 1300000))

    def test_original_imagesource_returned_when_no_callback(self):
        img_src1 = ImageSource(Image.new('RGBA', (100, 100)))
        env = make_wsgi_env('', extra_environ={})
        img_src2 = self.server.decorate_img(
            img_src1, 'wms.map', ['layer1'],
            env, self.query_extent
        )
        assert img_src1 == img_src2

    def test_returns_imagesource(self):
        img_src1 = ImageSource(Image.new('RGBA', (100, 100)))
        env = make_wsgi_env('', extra_environ={})
        img_src2 = self.server.decorate_img(
            img_src1, 'wms.map', ['layer1'],
            env, self.query_extent
        )
        assert isinstance(img_src2, ImageSource)

    def set_called_callback(self, img_src, service, layers, **kw):
        self.called = True
        return img_src

    def test_calls_callback(self):
        img_src1 = ImageSource(Image.new('RGBA', (100, 100)))
        self.called = False
        env = make_wsgi_env('', extra_environ={'mapproxy.decorate_img': self.set_called_callback})
        self.server.decorate_img(
            img_src1, 'wms.map', ['layer1'],
            env, self.query_extent
        )
        assert self.called is True

    def return_new_imagesource_callback(self, img_src, service, layers, **kw):
        new_img_src = ImageSource(Image.new('RGBA', (100, 100)))
        self.new_img_src = new_img_src
        return new_img_src

    def test_returns_callbacks_return_value(self):
        img_src1 = ImageSource(Image.new('RGBA', (100, 100)))
        env = make_wsgi_env('', extra_environ={'mapproxy.decorate_img': self.return_new_imagesource_callback})
        self.new_img_src = None
        img_src2 = self.server.decorate_img(
            img_src1, 'wms.map', ['layer1'],
            env, self.query_extent
        )
        assert img_src2 == self.new_img_src

    def test_wms_server(self):
        ''' Test that the decorate_img method is available on a WMSServer instance '''
        img_src1 = ImageSource(Image.new('RGBA', (100, 100)))
        self.called = False
        env = make_wsgi_env('', extra_environ={'mapproxy.decorate_img': self.set_called_callback})
        self.wms_server.decorate_img(
            img_src1, 'wms.map', ['layer1'],
            env, self.query_extent
        )
        assert self.called is True

    def test_tile_server(self):
        ''' Test that the decorate_img method is available on a TileServer instance '''
        img_src1 = ImageSource(Image.new('RGBA', (100, 100)))
        self.called = False
        env = make_wsgi_env('', extra_environ={'mapproxy.decorate_img': self.set_called_callback})
        self.tile_server.decorate_img(
            img_src1, 'tms', ['layer1'],
            env, self.query_extent
        )
        assert self.called is True

    def test_wmts_server(self):
        ''' Test that the decorate_img method is available on a WMTSServer instance '''
        img_src1 = ImageSource(Image.new('RGBA', (100, 100)))
        self.called = False
        env = make_wsgi_env('', extra_environ={'mapproxy.decorate_img': self.set_called_callback})
        self.wmts_server.decorate_img(
            img_src1, 'wmts', ['layer1'],
            env, self.query_extent
        )
        assert self.called is True

    def test_args(self):
        def callback(img_src, service, layers, environ, query_extent, **kw):
            assert isinstance(img_src, ImageSource)
            assert 'wms.map' == service
            assert isinstance(layers, list)
            assert isinstance(environ, dict)
            assert len(query_extent) == 2
            assert len(query_extent[1]) == 4
            return img_src
        img_src1 = ImageSource(Image.new('RGBA', (100, 100)))
        env = make_wsgi_env('', extra_environ={'mapproxy.decorate_img': callback})
        self.tile_server.decorate_img(
            img_src1, 'wms.map', ['layer1'],
            env, self.query_extent
        )



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO

from mapproxy.test.helper import Mocker
from mapproxy.test.mocker import ANY
from mapproxy.response import Response


class TestResponse(Mocker):

    def test_str_response(self):
        resp = Response("string content")
        assert isinstance(resp.response, str)
        start_response = self.mock()
        self.expect(start_response("200 OK", ANY))
        self.replay()
        result = resp({"REQUEST_METHOD": "GET"}, start_response)
        assert next(result) == b"string content"

    def test_itr_response(self):
        resp = Response(iter(["string content", "as iterable"]))
        assert hasattr(resp.response, "next") or hasattr(resp.response, "__next__")
        start_response = self.mock()
        self.expect(start_response("200 OK", ANY))
        self.replay()
        result = resp({"REQUEST_METHOD": "GET"}, start_response)
        assert next(result) == "string content"
        assert next(result) == "as iterable"

    def test_file_response(self):
        data = BytesIO(b"foobar")
        resp = Response(data)
        assert resp.response == data
        start_response = self.mock()
        self.expect(start_response("200 OK", ANY))
        self.replay()
        result = resp({"REQUEST_METHOD": "GET"}, start_response)
        assert next(result) == b"foobar"

    def test_file_response_w_file_wrapper(self):
        data = BytesIO(b"foobar")
        resp = Response(data)
        assert resp.response == data
        start_response = self.mock()
        self.expect(start_response("200 OK", ANY))

        file_wrapper = self.mock()
        self.expect(file_wrapper(data, resp.block_size)).result("DUMMY")
        self.replay()

        result = resp(
            {"REQUEST_METHOD": "GET", "wsgi.file_wrapper": file_wrapper}, start_response
        )
        assert result == "DUMMY"

    def test_file_response_content_length(self):
        data = BytesIO(b"*" * 342)
        resp = Response(data)
        assert resp.response == data
        start_response = self.mock()
        self.expect(start_response("200 OK", ANY))
        self.replay()
        resp({"REQUEST_METHOD": "GET"}, start_response)
        assert resp.content_length == 342
